# mypy: disable-error-code="attr-defined"
from opentelemetry import trace
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import SimpleSpanProcessor

from .exporter import _ConsoleExporter

TRACE_PROVIDER = trace.get_tracer_provider()
if isinstance(TRACE_PROVIDER, trace.ProxyTracerProvider):
    TRACE_PROVIDER = TracerProvider()
    trace.set_tracer_provider(TRACE_PROVIDER)


def enable_console_traces() -> None:
    """Enable printing traces to the console."""
    has_console_exporter = any(
        isinstance(getattr(p, "span_exporter", None), _ConsoleExporter)
        for p in TRACE_PROVIDER._active_span_processor._span_processors
    )
    if not has_console_exporter:
        TRACE_PROVIDER.add_span_processor(SimpleSpanProcessor(_ConsoleExporter()))


def disable_console_traces() -> None:
    """Disable printing traces to the console."""
    with TRACE_PROVIDER._active_span_processor._lock:
        TRACE_PROVIDER._active_span_processor._span_processors = tuple(
            p
            for p in TRACE_PROVIDER._active_span_processor._span_processors
            if not isinstance(getattr(p, "span_exporter", None), _ConsoleExporter)
        )


enable_console_traces()



from typing import TYPE_CHECKING, Any, cast

from pydantic import BaseModel

from any_agent.config import AgentConfig, AgentFramework

from .any_agent import AnyAgent

try:
    from langchain_core.language_models import LanguageModelLike
    from langchain_litellm import ChatLiteLLM

    # Patch the _OPENAI_MODELS list to include additional models
    # This can be removed after https://github.com/Akshay-Dongare/langchain-litellm/issues/7
    from langchain_litellm.chat_models import litellm as langchain_litellm_module
    from langgraph.prebuilt import create_react_agent

    # Add your custom models to the existing list
    additional_models = ["gpt-4.1-mini", "gpt-4.1-nano"]

    # Extend the existing _OPENAI_MODELS list
    if hasattr(langchain_litellm_module, "_OPENAI_MODELS"):
        langchain_litellm_module._OPENAI_MODELS.extend(additional_models)

    DEFAULT_AGENT_TYPE = create_react_agent
    DEFAULT_MODEL_TYPE = ChatLiteLLM

    langchain_available = True
except ImportError:
    langchain_available = False

if TYPE_CHECKING:
    from langchain_core.language_models import LanguageModelLike
    from langchain_core.messages.base import BaseMessage
    from langgraph.graph.graph import CompiledGraph


class LangchainAgent(AnyAgent):
    """LangChain agent implementation that handles both loading and running."""

    def __init__(self, config: AgentConfig):
        super().__init__(config)
        self._agent: CompiledGraph | None = None

    @property
    def framework(self) -> AgentFramework:
        return AgentFramework.LANGCHAIN

    def _get_model(self, agent_config: AgentConfig) -> "LanguageModelLike":
        """Get the model configuration for a LangChain agent."""
        model_type = agent_config.model_type or DEFAULT_MODEL_TYPE
        model_args = agent_config.model_args or {}
        return cast(
            "LanguageModelLike",
            model_type(
                model=agent_config.model_id,
                api_key=agent_config.api_key,
                api_base=agent_config.api_base,
                model_kwargs=model_args,  # type: ignore[arg-type]
            ),
        )

    async def _load_agent(self) -> None:
        """Load the LangChain agent with the given configuration."""
        if not langchain_available:
            msg = "You need to `pip install 'any-agent[langchain]'` to use this agent"
            raise ImportError(msg)

        imported_tools, _ = await self._load_tools(self.config.tools)

        self._tools = imported_tools
        agent_type = self.config.agent_type or DEFAULT_AGENT_TYPE
        agent_args = self.config.agent_args or {}
        if self.config.output_type:
            agent_args["response_format"] = self.config.output_type
        self._agent = agent_type(
            name=self.config.name,
            model=self._get_model(self.config),
            tools=imported_tools,
            prompt=self.config.instructions,
            **agent_args,
        )

    async def _run_async(self, prompt: str, **kwargs: Any) -> str | BaseModel:
        if not self._agent:
            error_message = "Agent not loaded. Call load_agent() first."
            raise ValueError(error_message)
        inputs = {"messages": [("user", prompt)]}
        result = await self._agent.ainvoke(inputs, **kwargs)
        if self.config.output_type:
            structured_response = result.get("structured_response")
            if not structured_response:
                msg = "No structured output returned from the agent."
                raise ValueError(msg)
            return structured_response  # type: ignore[no-any-return]
        if not result.get("messages"):
            msg = "No messages returned from the agent."
            raise ValueError(msg)
        last_message: BaseMessage = result["messages"][-1]
        return str(last_message.content)



from unittest.mock import MagicMock, patch

import pytest
from opentelemetry.sdk.trace import ReadableSpan

from any_agent import AgentTrace
from any_agent.tracing.exporter import _ConsoleExporter, _get_output_panel


@pytest.fixture
def readable_spans(agent_trace: AgentTrace) -> list[ReadableSpan]:
    return [span.to_readable_span() for span in agent_trace.spans]


def test_rich_console_span_exporter_default(
    agent_trace: AgentTrace, request: pytest.FixtureRequest
) -> None:
    console_mock = MagicMock()
    panel_mock = MagicMock()
    markdown_mock = MagicMock()
    readable_spans = [span.to_readable_span() for span in agent_trace.spans]
    with (
        patch("any_agent.tracing.exporter.Console", console_mock),
        patch("any_agent.tracing.exporter.Markdown", markdown_mock),
        patch("any_agent.tracing.exporter.Panel", panel_mock),
    ):
        exporter = _ConsoleExporter()
        exporter.export(readable_spans)
        console_mock.return_value.print.assert_called()
        # TINYAGENT ends with a `task_completed` tool call
        if request.node.callspec.id not in ("TINYAGENT_trace",):
            panel_mock.assert_any_call(
                markdown_mock(agent_trace.final_output),
                title="OUTPUT",
                style="white",
                title_align="left",
            )


def test_get_output_panel(
    readable_spans: list[ReadableSpan], request: pytest.FixtureRequest
) -> None:
    # First LLM call returns JSON
    panel_mock = MagicMock()
    json_mock = MagicMock()
    with (
        patch("any_agent.tracing.exporter.Panel", panel_mock),
        patch("any_agent.tracing.exporter.JSON", json_mock),
    ):
        _get_output_panel(readable_spans[0])
        json_mock.assert_called_once()
        panel_mock.assert_called_once()

    if request.node.callspec.id not in ("LLAMA_INDEX_trace",):
        # First TOOL execution returns JSON
        panel_mock = MagicMock()
        json_mock = MagicMock()
        with (
            patch("any_agent.tracing.exporter.Panel", panel_mock),
            patch("any_agent.tracing.exporter.JSON", json_mock),
        ):
            _get_output_panel(readable_spans[1])
            json_mock.assert_called_once()
            panel_mock.assert_called_once()

    if request.node.callspec.id not in ("TINYAGENT_trace",):
        # Final LLM call returns string
        panel_mock = MagicMock()
        json_mock = MagicMock()
        with (
            patch("any_agent.tracing.exporter.Panel", panel_mock),
            patch("any_agent.tracing.exporter.JSON", json_mock),
        ):
            _get_output_panel(readable_spans[-2])
            json_mock.assert_not_called()
            panel_mock.assert_called_once()

    # AGENT invocation has no output
    panel_mock = MagicMock()
    json_mock = MagicMock()
    with (
        patch("any_agent.tracing.exporter.Panel", panel_mock),
        patch("any_agent.tracing.exporter.JSON", json_mock),
    ):
        _get_output_panel(readable_spans[-1])
        json_mock.assert_not_called()
        panel_mock.assert_not_called()



from unittest.mock import MagicMock

from agents.tracing.span_data import GenerationSpanData

from any_agent.tracing.instrumentation.openai import (
    _OpenAIAgentsInstrumentor,
    _set_llm_input,
    _set_llm_output,
)


def test_set_llm_input_missing_fields() -> None:
    """It should not fail when missing fields."""
    span = MagicMock()
    _set_llm_input(GenerationSpanData(), span)

    span.set_attribute.assert_not_called()


def test_set_llm_output_missing_fields() -> None:
    """It should not fail when missing fields."""
    span = MagicMock()
    _set_llm_output(GenerationSpanData(), span)

    span.set_attributes.assert_not_called()


def test_uninstrument_before_instrument() -> None:
    _OpenAIAgentsInstrumentor().uninstrument(MagicMock())



import pytest
from litellm.utils import validate_environment

from any_agent import AgentConfig, AgentFramework, AnyAgent
from any_agent.tools import search_tavily


def test_no_instrument(
    agent_framework: AgentFramework,
) -> None:
    model_id = "gpt-4.1-nano"
    env_check = validate_environment("model_id")
    if not env_check["keys_in_environment"]:
        pytest.skip(f"{env_check['missing_keys']} needed for {agent_framework}")

    model_args = (
        {"parallel_tool_calls": False}
        if agent_framework not in [AgentFramework.AGNO, AgentFramework.LLAMA_INDEX]
        else None
    )

    agent = AnyAgent.create(
        agent_framework,
        AgentConfig(
            instructions="Use the available tools search the web for answers.",
            model_args=model_args,
            model_id=model_id,
            tools=[search_tavily],
        ),
    )

    assert agent._instrumentor

    agent_trace = agent.run(
        "How is the weather in Salvaterra de Miño?", instrument=False
    )

    assert not any(
        span.is_llm_call() or span.is_tool_execution() for span in agent_trace.spans
    )



import logging
from typing import Any

from rich.logging import RichHandler

logger = logging.getLogger("any_agent")


def setup_logger(
    level: int = logging.ERROR,
    rich_tracebacks: bool = True,
    log_format: str | None = None,
    propagate: bool = False,
    **kwargs: Any,
) -> None:
    """Configure the any_agent logger with the specified settings.

    Args:
        level: The logging level to use (default: logging.INFO)
        rich_tracebacks: Whether to enable rich tracebacks (default: True)
        log_format: Optional custom log format string
        propagate: Whether to propagate logs to parent loggers (default: False)
        **kwargs: Additional keyword arguments to pass to RichHandler

    """
    logger.setLevel(level)
    logger.propagate = propagate

    # Remove any existing handlers
    for handler in logger.handlers[:]:
        logger.removeHandler(handler)

    handler = RichHandler(rich_tracebacks=rich_tracebacks, markup=True, **kwargs)

    if log_format:
        formatter = logging.Formatter(log_format)
        handler.setFormatter(formatter)

    logger.addHandler(handler)


# Set default configuration
setup_logger()



from __future__ import annotations

import asyncio
import inspect
import json
from contextlib import suppress
from typing import TYPE_CHECKING, Any

import litellm
from litellm.utils import supports_response_schema
from mcp.types import CallToolResult, TextContent

from any_agent.config import AgentConfig, AgentFramework
from any_agent.logging import logger

from .any_agent import AnyAgent

if TYPE_CHECKING:
    from collections.abc import Callable

    from litellm.types.utils import Message as LiteLLMMessage
    from litellm.types.utils import ModelResponse
    from pydantic import BaseModel


DEFAULT_SYSTEM_PROMPT = """
You are an agent - please keep going until the user's query is completely resolved, before ending your turn and yielding back to the user. Only terminate your turn when you are sure that the problem is solved, or if you need more info from the user to solve the problem.

If you are not sure about anything pertaining to the user's request, use your tools to read files and gather the relevant information: do NOT guess or make up an answer.

You MUST plan extensively before each function call, and reflect extensively on the outcomes of the previous function calls. DO NOT do this entire process by making function calls only, as this can impair your ability to solve the problem and think insightfully.
""".strip()


DEFAULT_MAX_NUM_TURNS = 10


class ToolExecutor:
    """Executor for tools that wraps tool functions to work with the MCP client."""

    def __init__(self, tool_function: Callable[..., Any]) -> None:
        """Initialize the tool executor.

        Args:
            tool_function: The tool function to execute

        """
        self.tool_function = tool_function

    async def call_tool(self, request: dict[str, Any]) -> str:
        """Call the tool function.

        Args:
            request: The tool request with name and arguments

        Returns:
            Tool execution result

        """
        try:
            arguments = request.get("arguments", {})

            if hasattr(self.tool_function, "__annotations__"):
                func_args = self.tool_function.__annotations__
                for arg_name, arg_type in func_args.items():
                    if arg_name in arguments:
                        with suppress(Exception):
                            arguments[arg_name] = arg_type(arguments[arg_name])

            if asyncio.iscoroutinefunction(self.tool_function):
                result = await self.tool_function(**arguments)
            else:
                result = self.tool_function(**arguments)

            if (
                isinstance(result, CallToolResult)
                and result.content
                and isinstance(result.content[0], TextContent)
            ):
                result = result.content[0].text
            return str(result)

        except Exception as e:
            return f"Error executing tool: {e}"


class TinyAgent(AnyAgent):
    """A lightweight agent implementation using litellm.

    Modeled after JS implementation https://huggingface.co/blog/tiny-agents.
    """

    def __init__(self, config: AgentConfig) -> None:
        """Initialize the TinyAgent.

        Args:
            config: Agent configuration
            tracing: Optional tracing configuration

        """
        super().__init__(config)
        self.clients: dict[str, ToolExecutor] = {}
        self.completion_params = {
            "model": self.config.model_id,
            "tools": [],
            "tool_choice": "auto",
            **(self.config.model_args or {}),
        }
        if self.config.api_key:
            self.completion_params["api_key"] = self.config.api_key
        if self.config.api_base:
            self.completion_params["api_base"] = self.config.api_base

    async def _load_agent(self) -> None:
        """Load the agent and its tools."""
        wrapped_tools, mcp_servers = await self._load_tools(self.config.tools)
        self._mcp_servers = (
            mcp_servers  # Store servers so that they don't get garbage collected
        )

        self._tools = wrapped_tools

        for tool in wrapped_tools:
            tool_name = tool.__name__
            tool_desc = tool.__doc__ or f"Tool to {tool_name}"

            if not hasattr(tool, "__input_schema__"):
                sig = inspect.signature(tool)
                properties = {}
                required = []

                for param_name, param in sig.parameters.items():
                    if param.kind in (
                        inspect.Parameter.VAR_POSITIONAL,
                        inspect.Parameter.VAR_KEYWORD,
                    ):
                        continue

                    properties[param_name] = {
                        "type": "string",
                        "description": f"Parameter {param_name}",
                    }

                    if param.default == inspect.Parameter.empty:
                        required.append(param_name)

                input_schema = {
                    "type": "object",
                    "properties": properties,
                    "required": required,
                }
            else:
                input_schema = tool.__input_schema__

            self.completion_params["tools"].append(
                {
                    "type": "function",
                    "function": {
                        "name": tool_name,
                        "description": tool_desc,
                        "parameters": input_schema,
                    },
                }
            )

            self.clients[tool_name] = ToolExecutor(tool)
            logger.debug("Registered tool: %s", tool_name)

    async def _run_async(self, prompt: str, **kwargs: Any) -> str | BaseModel:
        messages = [
            {
                "role": "system",
                "content": self.config.instructions or DEFAULT_SYSTEM_PROMPT,
            },
            {
                "role": "user",
                "content": prompt,
            },
        ]

        num_of_turns = 0
        max_turns = kwargs.get("max_turns", DEFAULT_MAX_NUM_TURNS)

        while num_of_turns < max_turns:
            self.completion_params["messages"] = messages
            response = await self.call_model(**self.completion_params)

            message: LiteLLMMessage = response.choices[0].message  # type: ignore[union-attr]

            messages.append(message.model_dump())

            if message.tool_calls:
                for tool_call in message.tool_calls:
                    tool_name = tool_call.function.name
                    tool_message = {
                        "role": "tool",
                        "tool_call_id": tool_call.id,
                        "content": "",
                        "name": tool_name,
                    }

                    if tool_name not in self.clients:
                        tool_message["content"] = (
                            f"Error: No tool found with name: {tool_name}"
                        )
                        continue

                    tool_args = {}
                    if tool_call.function.arguments:
                        tool_args = json.loads(tool_call.function.arguments)

                    client = self.clients[tool_name]
                    result = await client.call_tool(
                        {"name": tool_name, "arguments": tool_args}
                    )
                    tool_message["content"] = result
                    messages.append(tool_message)  # type: ignore[arg-type]

            num_of_turns += 1
            current_last = messages[-1]
            if current_last.get("role") == "assistant" and current_last.get("content"):
                if self.config.output_type:
                    structured_output_message = {
                        "role": "user",
                        "content": f"Please conform your output to the following schema: {self.config.output_type.model_json_schema()}.",
                    }
                    messages.append(structured_output_message)
                    self.completion_params["messages"] = messages
                    if supports_response_schema(model=self.config.model_id):
                        self.completion_params["response_format"] = (
                            self.config.output_type
                        )
                    response = await litellm.acompletion(**self.completion_params)
                    return self.config.output_type.model_validate_json(
                        response.choices[0].message["content"]  # type: ignore[union-attr]
                    )
                return str(current_last["content"])

        return "Max turns reached"

    async def call_model(self, **completion_params: dict[str, Any]) -> ModelResponse:
        response: ModelResponse = await litellm.acompletion(**completion_params)
        return response

    @property
    def framework(self) -> AgentFramework:
        return AgentFramework.TINYAGENT



from abc import ABC, abstractmethod
from collections.abc import Sequence
from contextlib import suppress
from datetime import timedelta
from typing import Any, Literal

from pydantic import Field, PrivateAttr

from any_agent.config import AgentFramework, MCPSse, MCPStdio
from any_agent.tools.mcp.mcp_connection import _MCPConnection
from any_agent.tools.mcp.mcp_server import _MCPServerBase

mcp_available = False
with suppress(ImportError):
    from langchain_core.tools import BaseTool  # noqa: TC002
    from langchain_mcp_adapters.tools import load_mcp_tools
    from mcp import ClientSession, StdioServerParameters
    from mcp.client.sse import sse_client
    from mcp.client.stdio import stdio_client

    mcp_available = True


class LangchainMCPConnection(_MCPConnection["BaseTool"], ABC):
    """Base class for LangChain MCP connections."""

    _client: Any | None = PrivateAttr(default=None)

    @abstractmethod
    async def list_tools(self) -> list["BaseTool"]:
        """List tools from the MCP server."""
        if not self._client:
            msg = "MCP client is not set up. Please call `list_tools` from a concrete class."
            raise ValueError(msg)

        stdio, write = await self._exit_stack.enter_async_context(self._client)
        kwargs = {}
        if self.mcp_tool.client_session_timeout_seconds:
            kwargs["read_timeout_seconds"] = timedelta(
                seconds=self.mcp_tool.client_session_timeout_seconds
            )
        client_session = ClientSession(
            stdio,
            write,
            **kwargs,  # type: ignore[arg-type]
        )
        session = await self._exit_stack.enter_async_context(client_session)

        await session.initialize()

        tools = await load_mcp_tools(session)
        return self._filter_tools(tools)  # type: ignore[return-value]


class LangchainMCPStdioConnection(LangchainMCPConnection):
    mcp_tool: MCPStdio

    async def list_tools(self) -> list["BaseTool"]:
        """List tools from the MCP server."""
        server_params = StdioServerParameters(
            command=self.mcp_tool.command,
            args=list(self.mcp_tool.args),
            env=self.mcp_tool.env,
        )

        self._client = stdio_client(server_params)

        return await super().list_tools()


class LangchainMCPSseConnection(LangchainMCPConnection):
    mcp_tool: MCPSse

    async def list_tools(self) -> list["BaseTool"]:
        """List tools from the MCP server."""
        kwargs = {}
        if self.mcp_tool.client_session_timeout_seconds:
            kwargs["sse_read_timeout"] = self.mcp_tool.client_session_timeout_seconds
        self._client = sse_client(
            url=self.mcp_tool.url,
            headers=dict(self.mcp_tool.headers or {}),
            **kwargs,  # type: ignore[arg-type]
        )
        return await super().list_tools()


class LangchainMCPServerBase(_MCPServerBase["BaseTool"], ABC):
    framework: Literal[AgentFramework.LANGCHAIN] = AgentFramework.LANGCHAIN
    tools: Sequence["BaseTool"] = Field(default_factory=list)

    def _check_dependencies(self) -> None:
        self.libraries = "any-agent[mcp,langchain]"
        self.mcp_available = mcp_available
        super()._check_dependencies()


class LangchainMCPServerStdio(LangchainMCPServerBase):
    mcp_tool: MCPStdio

    async def _setup_tools(
        self, mcp_connection: _MCPConnection["BaseTool"] | None = None
    ) -> None:
        mcp_connection = mcp_connection or LangchainMCPStdioConnection(
            mcp_tool=self.mcp_tool
        )
        await super()._setup_tools(mcp_connection)


class LangchainMCPServerSse(LangchainMCPServerBase):
    mcp_tool: MCPSse

    async def _setup_tools(
        self, mcp_connection: _MCPConnection["BaseTool"] | None = None
    ) -> None:
        mcp_connection = mcp_connection or LangchainMCPSseConnection(
            mcp_tool=self.mcp_tool
        )
        await super()._setup_tools(mcp_connection)


LangchainMCPServer = LangchainMCPServerStdio | LangchainMCPServerSse



import pytest
from pydantic import BaseModel

# Skip entire module if a2a dependencies are not available
pytest.importorskip("a2a.types")
pytest.importorskip("any_agent.serving.envelope")

from a2a.types import TaskState

from any_agent.config import AgentConfig, AgentFramework
from any_agent.frameworks.any_agent import AnyAgent
from any_agent.serving.a2a.envelope import (
    A2AEnvelope,
    _DefaultBody,
    _is_a2a_envelope,
    prepare_agent_for_a2a,
)


class CustomOutputType(BaseModel):
    custom_field: str
    result: str


class MockAgent(AnyAgent):
    """Mock agent implementation for testing."""

    def __init__(self, config: AgentConfig) -> None:
        super().__init__(config)
        self._agent = None

    async def _load_agent(self) -> None:
        pass

    async def _run_async(self, prompt: str, **kwargs: object) -> str:
        return "mock result"

    @property
    def framework(self) -> AgentFramework:
        from any_agent.config import AgentFramework

        return AgentFramework.TINYAGENT

    @classmethod
    def create(cls, framework: object, config: AgentConfig) -> "MockAgent":
        return cls(config)


def test_envelope_created_without_output_type() -> None:
    """Test that the envelope is correctly created when the agent is configured without an output_type."""
    # Create agent config without output_type
    config = AgentConfig(model_id="test-model", description="test agent")
    assert config.output_type is None

    # Create mock agent
    agent = MockAgent(config)

    # Prepare agent for A2A
    prepared_agent = prepare_agent_for_a2a(agent)

    # Verify the envelope was created with default body
    assert prepared_agent.config.output_type is not None
    assert _is_a2a_envelope(prepared_agent.config.output_type)

    # Verify the envelope wraps _DefaultBody
    envelope_instance = prepared_agent.config.output_type(
        task_status=TaskState.completed, data=_DefaultBody(result="test result")
    )

    assert isinstance(envelope_instance, A2AEnvelope)
    assert envelope_instance.task_status == TaskState.completed
    assert isinstance(envelope_instance.data, _DefaultBody)
    assert envelope_instance.data.result == "test result"


def test_envelope_created_with_output_type() -> None:
    """Test that the envelope is correctly created when an agent is configured with an output_type."""
    # Create agent config with custom output_type
    config = AgentConfig(
        model_id="test-model", description="test agent", output_type=CustomOutputType
    )

    # Create mock agent
    agent = MockAgent(config)

    # Prepare agent for A2A
    prepared_agent = prepare_agent_for_a2a(agent)

    # Verify the envelope was created with custom output type
    assert prepared_agent.config.output_type is not None
    assert _is_a2a_envelope(prepared_agent.config.output_type)

    # Verify the envelope wraps the custom output type
    envelope_instance = prepared_agent.config.output_type(
        task_status=TaskState.completed,
        data=CustomOutputType(custom_field="test", result="custom result"),
    )

    assert isinstance(envelope_instance, A2AEnvelope)
    assert envelope_instance.task_status == TaskState.completed
    assert isinstance(envelope_instance.data, CustomOutputType)
    assert envelope_instance.data.custom_field == "test"
    assert envelope_instance.data.result == "custom result"



from any_agent.tracing.agent_trace import AgentSpan, AgentTrace
from any_agent.tracing.otel_types import Resource, SpanContext, SpanKind, Status


def create_llm_span(input_tokens: int = 100, output_tokens: int = 50) -> AgentSpan:
    """Create a mock LLM span with token usage."""
    return AgentSpan(
        name="call_llm gpt-4o-mini",
        kind=SpanKind.INTERNAL,
        status=Status(),
        context=SpanContext(span_id=123),
        attributes={
            "gen_ai.operation.name": "call_llm",
            "gen_ai.usage.input_tokens": input_tokens,
            "gen_ai.usage.output_tokens": output_tokens,
        },
        links=[],
        events=[],
        resource=Resource(),
    )


def test_tokens_and_cost_properties_are_cached() -> None:
    """Test that tokens and cost properties are cached after first access."""
    trace = AgentTrace()
    trace.add_span(create_llm_span(input_tokens=100, output_tokens=50))

    # First access - should compute and cache
    tokens1 = trace.tokens
    cost1 = trace.cost

    # Second access - should return cached objects
    tokens2 = trace.tokens
    cost2 = trace.cost

    assert tokens1 is tokens2  # Same object reference indicates caching
    assert cost1 is cost2  # Same object reference indicates caching
    assert "tokens" in trace.__dict__
    assert "cost" in trace.__dict__


def test_add_span_invalidates_cache() -> None:
    """Test that adding a span invalidates both tokens and cost caches."""
    trace = AgentTrace()
    trace.add_span(create_llm_span(input_tokens=100, output_tokens=50))

    # Cache the properties
    _ = trace.tokens
    _ = trace.cost
    assert "tokens" in trace.__dict__
    assert "cost" in trace.__dict__

    # Add another span - should invalidate cache
    trace.add_span(create_llm_span(input_tokens=200, output_tokens=75))

    assert "tokens" not in trace.__dict__
    assert "cost" not in trace.__dict__

    # Verify new calculations are correct
    tokens = trace.tokens
    assert tokens.input_tokens == 300
    assert tokens.output_tokens == 125


def test_invalidate_cache_method() -> None:
    """Test that _invalidate_tokens_and_cost_cache clears both caches."""
    trace = AgentTrace()
    trace.add_span(create_llm_span())

    # Cache both properties
    _ = trace.tokens
    _ = trace.cost
    assert "tokens" in trace.__dict__
    assert "cost" in trace.__dict__

    # Manually invalidate cache
    trace._invalidate_tokens_and_cost_cache()

    assert "tokens" not in trace.__dict__
    assert "cost" not in trace.__dict__


def test_spans_to_messages_handles_empty_spans() -> None:
    """Test that spans_to_messages handles traces with no spans."""
    empty_trace = AgentTrace()
    messages = empty_trace.spans_to_messages()

    assert isinstance(messages, list)
    assert len(messages) == 0



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from any_agent import AgentFramework
from any_agent.config import AgentConfig, MCPSse, MCPStdio, Tool
from any_agent.frameworks.any_agent import AnyAgent
from any_agent.tools import _get_mcp_server

if TYPE_CHECKING:
    from collections.abc import Generator, Sequence

    from agents.mcp import MCPServerSse as OpenAIInternalMCPServerSse


@pytest.fixture
def openai_mcp_sse_server(
    tools: Sequence[Tool],
) -> Generator[OpenAIInternalMCPServerSse]:
    with patch(
        "any_agent.tools.mcp.frameworks.openai.OpenAIInternalMCPServerSse",
    ) as mock_server:
        mock_server.return_value.list_tools = AsyncMock(return_value=tools)
        yield mock_server


@pytest.mark.usefixtures(
    "openai_mcp_sse_server",
)
def test_openai_mcpsse(
    mcp_sse_params_no_tools: MCPSse,
) -> None:
    """This is a test kept for legacy purposes."""
    agent_config = AgentConfig(model_id="gpt-4o", tools=[mcp_sse_params_no_tools])

    agent = AnyAgent.create("openai", agent_config)

    servers = agent._mcp_servers
    assert servers

    server, *_ = agent._mcp_servers
    assert server.mcp_tool == mcp_sse_params_no_tools


@pytest.mark.asyncio
async def test_openai_mcp_env() -> None:
    mcp_server = _get_mcp_server(
        MCPStdio(command="print('Hello MCP')", args=[], env={"FOO": "BAR"}),
        AgentFramework.OPENAI,
    )
    mocked_class = MagicMock()
    mocked_class.return_value = AsyncMock()

    with (
        patch(
            "any_agent.tools.mcp.frameworks.openai.OpenAIInternalMCPServerStdio",
            mocked_class,
        ),
    ):
        await mcp_server._setup_tools()
        assert mocked_class.call_args_list[0][1]["params"]["env"] == {"FOO": "BAR"}


def test_openai_client_session_timeout_passed() -> None:
    """Test that client_session_timeout_seconds parameter is properly passed to OpenAI MCPServerStdio and MCPServerSse."""
    custom_timeout = 15.0
    stdio_params = MCPStdio(
        command="echo",
        args=["test"],
        client_session_timeout_seconds=custom_timeout,
    )
    sse_params = MCPSse(
        url="http://localhost:8000",
        client_session_timeout_seconds=custom_timeout,
    )
    # STDIO
    server = _get_mcp_server(stdio_params, AgentFramework.OPENAI)
    with patch(
        "any_agent.tools.mcp.frameworks.openai.OpenAIInternalMCPServerStdio"
    ) as mock_stdio:
        mock_server_instance = AsyncMock()
        mock_server_instance.__aenter__ = AsyncMock(return_value=mock_server_instance)
        mock_server_instance.__aexit__ = AsyncMock(return_value=None)
        mock_server_instance.list_tools = AsyncMock(return_value=[])
        mock_stdio.return_value = mock_server_instance
        import asyncio

        asyncio.run(server._setup_tools())
        mock_stdio.assert_called_once()
        call_args = mock_stdio.call_args
        assert call_args.kwargs["client_session_timeout_seconds"] == custom_timeout
    # SSE
    server = _get_mcp_server(sse_params, AgentFramework.OPENAI)
    with patch(
        "any_agent.tools.mcp.frameworks.openai.OpenAIInternalMCPServerSse"
    ) as mock_sse:
        mock_server_instance = AsyncMock()
        mock_server_instance.__aenter__ = AsyncMock(return_value=mock_server_instance)
        mock_server_instance.__aexit__ = AsyncMock(return_value=None)
        mock_server_instance.list_tools = AsyncMock(return_value=[])
        mock_sse.return_value = mock_server_instance
        asyncio.run(server._setup_tools())
        mock_sse.assert_called_once()
        call_args = mock_sse.call_args
        assert call_args.kwargs["client_session_timeout_seconds"] == custom_timeout



from typing import TYPE_CHECKING, Any

from pydantic import BaseModel

from any_agent.config import AgentConfig, AgentFramework
from any_agent.frameworks.any_agent import AnyAgent

try:
    from smolagents import FinalAnswerTool, LiteLLMModel, ToolCallingAgent

    smolagents_available = True
except ImportError:
    smolagents_available = False

if TYPE_CHECKING:
    from smolagents import MultiStepAgent


DEFAULT_AGENT_TYPE = ToolCallingAgent
DEFAULT_MODEL_TYPE = LiteLLMModel


class SmolagentsAgent(AnyAgent):
    """Smolagents agent implementation that handles both loading and running."""

    def __init__(
        self,
        config: AgentConfig,
    ):
        super().__init__(config)
        self._agent: MultiStepAgent | None = None

    @property
    def framework(self) -> AgentFramework:
        return AgentFramework.SMOLAGENTS

    def _get_model(self, agent_config: AgentConfig) -> Any:
        """Get the model configuration for a smolagents agent."""
        model_type = agent_config.model_type or DEFAULT_MODEL_TYPE
        model_args = agent_config.model_args or {}
        kwargs = {
            "model_id": agent_config.model_id,
            "api_key": agent_config.api_key,
            "api_base": agent_config.api_base,
            **model_args,
        }
        return model_type(**kwargs)

    async def _load_agent(self) -> None:
        """Load the Smolagents agent with the given configuration."""
        if not smolagents_available:
            msg = "You need to `pip install 'any-agent[smolagents]'` to use this agent"
            raise ImportError(msg)

        tools, _ = await self._load_tools(self.config.tools)

        main_agent_type = self.config.agent_type or DEFAULT_AGENT_TYPE

        agent_args = self.config.agent_args or {}

        self._tools = tools
        self._agent = main_agent_type(
            name=self.config.name,
            model=self._get_model(self.config),
            tools=tools,
            verbosity_level=-1,  # OFF
            **agent_args,
        )

        if self.config.instructions:
            self._agent.prompt_templates["system_prompt"] = self.config.instructions

        if self.config.output_type:
            output_type = self.config.output_type

            class CustomFinalAnswerTool(FinalAnswerTool):  # type: ignore[no-untyped-call]
                inputs = {  # noqa: RUF012
                    "answer": {
                        "type": "string",
                        "description": f"The final answer to the problem. The input must be a string that conforms to the{output_type.__name__} object.",
                    }
                }

                def forward(self, answer: str) -> Any:
                    output_type.model_validate_json(answer)
                    return answer

            self._agent.tools["final_answer"] = CustomFinalAnswerTool()  # type: ignore[no-untyped-call]

            self._agent.prompt_templates[
                "system_prompt"
            ] += f"""\n\nYour final answer must be a {self.config.output_type.__name__} object.
            This object must match the following schema:
            {self.config.output_type.model_json_schema()}
            """
        assert self._agent

    async def _run_async(self, prompt: str, **kwargs: Any) -> str | BaseModel:
        if not self._agent:
            error_message = "Agent not loaded. Call load_agent() first."
            raise ValueError(error_message)
        result = self._agent.run(prompt, **kwargs)
        if self.config.output_type:
            return self.config.output_type.model_validate_json(result)
        return str(result)



from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import Literal

from pydantic import Field, PrivateAttr

from any_agent.config import (
    AgentFramework,
    MCPSse,
    MCPStdio,
)
from any_agent.tools.mcp.mcp_connection import _MCPConnection
from any_agent.tools.mcp.mcp_server import _MCPServerBase

try:
    from google.adk.tools.mcp_tool import MCPTool as GoogleMCPTool
    from google.adk.tools.mcp_tool import MCPToolset as GoogleMCPToolset
    from google.adk.tools.mcp_tool.mcp_session_manager import (
        SseConnectionParams as GoogleSseServerParameters,
    )
    from mcp import (
        StdioServerParameters as GoogleStdioServerParameters,
    )

    mcp_available = True
except ImportError:
    mcp_available = False


class GoogleMCPConnection(_MCPConnection["GoogleMCPTool"], ABC):
    """Base class for Google MCP connections."""

    _params: "GoogleStdioServerParameters | GoogleSseServerParameters | None" = (
        PrivateAttr(default=None)
    )

    @abstractmethod
    async def list_tools(self) -> list["GoogleMCPTool"]:
        """List tools from the MCP server."""
        if not self._params:
            msg = "MCP params is not set up. Please call `list_tools` from a concrete class."
            raise ValueError(msg)

        server = GoogleMCPToolset(connection_params=self._params)
        tools = await server.get_tools()
        return self._filter_tools(tools)  # type: ignore[return-value]


class GoogleMCPStdioConnection(GoogleMCPConnection):
    mcp_tool: MCPStdio

    async def list_tools(self) -> list["GoogleMCPTool"]:
        """List tools from the MCP server."""
        self._params = GoogleStdioServerParameters(
            command=self.mcp_tool.command,
            args=list(self.mcp_tool.args),
            env=self.mcp_tool.env,
        )
        return await super().list_tools()


class GoogleMCPSseConnection(GoogleMCPConnection):
    mcp_tool: MCPSse

    async def list_tools(self) -> list["GoogleMCPTool"]:
        """List tools from the MCP server."""
        self._params = GoogleSseServerParameters(
            url=self.mcp_tool.url,
            headers=dict(self.mcp_tool.headers or {}),
        )
        return await super().list_tools()


class GoogleMCPServerBase(_MCPServerBase["GoogleMCPTool"], ABC):
    framework: Literal[AgentFramework.GOOGLE] = AgentFramework.GOOGLE
    tools: Sequence["GoogleMCPTool"] = Field(default_factory=list)

    def _check_dependencies(self) -> None:
        """Check if the required dependencies for the MCP server are available."""
        self.libraries = "any-agent[mcp,google]"
        self.mcp_available = mcp_available
        super()._check_dependencies()


class GoogleMCPServerStdio(GoogleMCPServerBase):
    mcp_tool: MCPStdio

    async def _setup_tools(
        self, mcp_connection: _MCPConnection["GoogleMCPTool"] | None = None
    ) -> None:
        mcp_connection = mcp_connection or GoogleMCPStdioConnection(
            mcp_tool=self.mcp_tool
        )
        await super()._setup_tools(mcp_connection)


class GoogleMCPServerSse(GoogleMCPServerBase):
    mcp_tool: MCPSse

    async def _setup_tools(
        self, mcp_connection: _MCPConnection["GoogleMCPTool"] | None = None
    ) -> None:
        mcp_connection = mcp_connection or GoogleMCPSseConnection(
            mcp_tool=self.mcp_tool
        )
        await super()._setup_tools(mcp_connection)


GoogleMCPServer = GoogleMCPServerStdio | GoogleMCPServerSse



from unittest.mock import ANY, AsyncMock, MagicMock, patch

import pytest

from any_agent import AgentConfig, AgentFramework, AnyAgent


def test_load_langchain_agent_default() -> None:
    model_mock = MagicMock()
    create_mock = MagicMock()
    agent_mock = MagicMock()
    create_mock.return_value = agent_mock
    tool_mock = MagicMock()

    with (
        patch("any_agent.frameworks.langchain.DEFAULT_AGENT_TYPE", create_mock),
        patch("any_agent.frameworks.langchain.DEFAULT_MODEL_TYPE", model_mock),
        patch("langchain_core.tools.tool", tool_mock),
    ):
        AnyAgent.create(AgentFramework.LANGCHAIN, AgentConfig(model_id="gpt-4o"))

        model_mock.assert_called_once_with(
            model="gpt-4o", api_base=None, api_key=None, model_kwargs={}
        )
        create_mock.assert_called_once_with(
            name="any_agent",
            model=model_mock.return_value,
            tools=[],
            prompt=None,
        )


def test_load_langchain_agent_missing() -> None:
    with patch("any_agent.frameworks.langchain.langchain_available", False):
        with pytest.raises(ImportError):
            AnyAgent.create(AgentFramework.LANGCHAIN, AgentConfig(model_id="gpt-4o"))


def test_run_langchain_agent_custom_args() -> None:
    create_mock = MagicMock()
    agent_mock = AsyncMock()
    agent_mock.ainvoke.return_value = MagicMock()
    create_mock.return_value = agent_mock

    with (
        patch("any_agent.frameworks.langchain.DEFAULT_AGENT_TYPE", create_mock),
        patch("any_agent.frameworks.langchain.DEFAULT_MODEL_TYPE"),
        patch("langchain_core.tools.tool"),
    ):
        agent = AnyAgent.create(
            AgentFramework.LANGCHAIN, AgentConfig(model_id="gpt-4o")
        )
        agent.run("foo", debug=True)
        agent_mock.ainvoke.assert_called_once_with(
            {"messages": [("user", "foo")]}, debug=True, config={"callbacks": [ANY]}
        )



import json

from pydantic import BaseModel, ValidationError


class FinalOutputTool:
    """A serializable final output tool that avoids closure issues."""

    def __init__(self, output_type: type[BaseModel]):
        """Create the function that will be used as a tool."""
        self.output_type = output_type
        # Set docstring for the callable object
        self.__doc__ = f"""You must call this tool in order to return the final answer.

        Args:
            answer: The final output that can be loaded as a Pydantic model. This must be a JSON compatible string that matches the following schema:
                {output_type.model_json_schema()}

        Returns:
            A dictionary with the following keys:
                - success: True if the output is valid, False otherwise.
                - result: The final output if success is True, otherwise an error message.

        """
        # Set function name for tool frameworks that expect it
        self.__name__ = "final_output"

    def __call__(self, answer: str) -> dict:  # type: ignore[type-arg]
        """Validate the final output."""
        # First check if it's valid JSON
        try:
            json.loads(answer)
        except json.JSONDecodeError as json_err:
            return {
                "success": False,
                "result": f"Invalid JSON format: {json_err}. Please fix the 'answer' parameter so that it is a valid JSON string and call this tool again.",
            }
        # Then validate against the Pydantic model
        try:
            self.output_type.model_validate_json(answer)
        except ValidationError as e:
            return {
                "success": False,
                "result": f"Please fix this validation error: {e}. The format must conform to {self.output_type.model_json_schema()}",
            }
        else:
            return {"success": True, "result": answer}



import json
from unittest.mock import MagicMock, patch

import pytest
from opentelemetry.trace import StatusCode

from any_agent.tracing.instrumentation.common import _set_tool_output


class FooClass:
    pass


foo_instance = FooClass()


@pytest.mark.parametrize(
    ("tool_output", "expected_output", "expected_output_type"),
    [
        ("foo", "foo", "text"),
        (json.dumps({"foo": "bar"}), json.dumps({"foo": "bar"}), "json"),
        ({"foo": "bar"}, json.dumps({"foo": "bar"}), "json"),
        (foo_instance, json.dumps(foo_instance, default=str), "json"),
    ],
)
def test_set_tool_output(
    tool_output: object, expected_output: str, expected_output_type: str
) -> None:
    span_mock = MagicMock()
    _set_tool_output(tool_output, span_mock)

    span_mock.set_attributes.assert_called_with(
        {"gen_ai.output": expected_output, "gen_ai.output.type": expected_output_type}
    )
    span_mock.set_status.assert_called_with(StatusCode.OK)


def test_set_tool_output_error() -> None:
    error = "Error calling tool: It's a trap!"
    span_mock = MagicMock()
    status_mock = MagicMock()
    with patch("any_agent.tracing.instrumentation.common.Status", status_mock):
        _set_tool_output(error, span_mock)

        span_mock.set_attributes.assert_called_with(
            {"gen_ai.output": error, "gen_ai.output.type": "text"}
        )
        span_mock.set_status.assert_called_with(
            status_mock(status_code=StatusCode.ERROR, description=error)
        )



from unittest.mock import MagicMock

from llama_index.core.agent.workflow.workflow_events import AgentOutput
from llama_index.core.base.llms.types import ChatMessage

from any_agent.tracing.instrumentation.llama_index import (
    _LlamaIndexInstrumentor,
    _set_llm_input,
    _set_llm_output,
)


def test_set_llm_input_missing_fields() -> None:
    """It should not fail when missing fields."""
    span = MagicMock()
    _set_llm_input([ChatMessage()], span)

    span.set_attribute.assert_called_with(
        "gen_ai.input.messages", '[{"role": "user", "content": "No content"}]'
    )


def test_set_llm_output_missing_fields() -> None:
    """It should not fail when missing fields."""
    span = MagicMock()
    _set_llm_output(
        AgentOutput(
            response=ChatMessage(), tool_calls=[], raw=None, current_agent_name="foo"
        ),
        span,
    )

    span.set_attributes.assert_not_called()


def test_uninstrument_before_instrument() -> None:
    _LlamaIndexInstrumentor().uninstrument(MagicMock())



# pylint: disable=unused-argument, unused-variable, attr-de
from collections.abc import Generator, Sequence
from unittest.mock import MagicMock, patch

import pytest
from smolagents.mcp_client import MCPClient

from any_agent.config import AgentFramework, MCPSse, MCPStdio, Tool
from any_agent.tools import _get_mcp_server


@pytest.fixture
def smolagents_mcp_server(
    tools: Sequence[Tool],
) -> Generator[MCPClient]:
    with patch(
        "any_agent.tools.mcp.frameworks.smolagents.MCPClient"
    ) as mock_client_class:
        mock_client_class.return_value.__enter__.return_value = tools
        yield mock_client_class


@pytest.mark.asyncio
async def test_smolagents_mcp_sse_integration(
    mcp_sse_params_no_tools: MCPSse,
    smolagents_mcp_server: MCPClient,
) -> None:
    server = _get_mcp_server(mcp_sse_params_no_tools, AgentFramework.SMOLAGENTS)

    await server._setup_tools()

    smolagents_mcp_server.assert_called_once_with({"url": mcp_sse_params_no_tools.url})  # type: ignore[attr-defined]


@pytest.mark.asyncio
async def test_smolagents_mcp_env() -> None:
    mcp_server = _get_mcp_server(
        MCPStdio(command="print('Hello MCP')", args=[], env={"FOO": "BAR"}),
        AgentFramework.SMOLAGENTS,
    )
    mocked_class = MagicMock()

    with (
        patch("any_agent.tools.mcp.frameworks.smolagents.MCPClient", mocked_class),
    ):
        await mcp_server._setup_tools()
        assert mocked_class.call_args_list[0][0][0].env == {"FOO": "BAR"}



from abc import ABC, abstractmethod
from collections.abc import Sequence
from contextlib import suppress
from typing import Literal

from pydantic import Field, PrivateAttr

from any_agent.config import AgentFramework, MCPSse, MCPStdio
from any_agent.tools.mcp.mcp_connection import _MCPConnection
from any_agent.tools.mcp.mcp_server import _MCPServerBase

mcp_available = False
with suppress(ImportError):
    from llama_index.core.tools import (
        FunctionTool as LlamaIndexFunctionTool,  # noqa: TC002
    )
    from llama_index.tools.mcp import BasicMCPClient as LlamaIndexMCPClient
    from llama_index.tools.mcp import McpToolSpec as LlamaIndexMcpToolSpec

    mcp_available = True


class LlamaIndexMCPConnection(_MCPConnection["LlamaIndexFunctionTool"], ABC):
    """Base class for LlamaIndex MCP connections."""

    _client: "LlamaIndexMCPClient | None" = PrivateAttr(default=None)

    @abstractmethod
    async def list_tools(self) -> list["LlamaIndexFunctionTool"]:
        """List tools from the MCP server."""
        if not self._client:
            msg = "MCP client is not set up. Please call `list_tool` from a concrete class."
            raise ValueError(msg)

        allowed_tools = self.mcp_tool.tools
        if allowed_tools is not None:
            allowed_tools = list(allowed_tools)
        mcp_tool_spec = LlamaIndexMcpToolSpec(
            client=self._client,
            allowed_tools=allowed_tools,
        )

        return await mcp_tool_spec.to_tool_list_async()


class LlamaIndexMCPStdioConnection(LlamaIndexMCPConnection):
    mcp_tool: MCPStdio

    async def list_tools(self) -> list["LlamaIndexFunctionTool"]:
        """List tools from the MCP server."""
        kwargs = {}
        if self.mcp_tool.client_session_timeout_seconds:
            kwargs["timeout"] = self.mcp_tool.client_session_timeout_seconds
        self._client = LlamaIndexMCPClient(
            command_or_url=self.mcp_tool.command,
            args=list(self.mcp_tool.args),
            env=self.mcp_tool.env,
            **kwargs,  # type: ignore[arg-type]
        )
        return await super().list_tools()


class LlamaIndexMCPSseConnection(LlamaIndexMCPConnection):
    mcp_tool: MCPSse

    async def list_tools(self) -> list["LlamaIndexFunctionTool"]:
        """List tools from the MCP server."""
        kwargs = {}
        if self.mcp_tool.client_session_timeout_seconds:
            kwargs["timeout"] = self.mcp_tool.client_session_timeout_seconds
        self._client = LlamaIndexMCPClient(
            command_or_url=self.mcp_tool.url,
            **kwargs,  # type: ignore[arg-type]
        )
        return await super().list_tools()


class LlamaIndexMCPServerBase(_MCPServerBase["LlamaIndexFunctionTool"], ABC):
    framework: Literal[AgentFramework.LLAMA_INDEX] = AgentFramework.LLAMA_INDEX
    tools: Sequence["LlamaIndexFunctionTool"] = Field(default_factory=list)

    def _check_dependencies(self) -> None:
        """Check if the required dependencies for the MCP server are available."""
        self.libraries = "any-agent[mcp,llama_index]"
        self.mcp_available = mcp_available
        super()._check_dependencies()


class LlamaIndexMCPServerStdio(LlamaIndexMCPServerBase):
    mcp_tool: MCPStdio

    async def _setup_tools(
        self, mcp_connection: _MCPConnection["LlamaIndexFunctionTool"] | None = None
    ) -> None:
        mcp_connection = mcp_connection or LlamaIndexMCPStdioConnection(
            mcp_tool=self.mcp_tool
        )
        await super()._setup_tools(mcp_connection)


class LlamaIndexMCPServerSse(LlamaIndexMCPServerBase):
    mcp_tool: MCPSse

    async def _setup_tools(
        self, mcp_connection: _MCPConnection["LlamaIndexFunctionTool"] | None = None
    ) -> None:
        mcp_connection = mcp_connection or LlamaIndexMCPSseConnection(
            mcp_tool=self.mcp_tool
        )
        await super()._setup_tools(mcp_connection)


LlamaIndexMCPServer = LlamaIndexMCPServerStdio | LlamaIndexMCPServerSse



from multiprocessing import Process, Queue
from typing import TYPE_CHECKING, Any

import pytest
from litellm.utils import validate_environment

from any_agent import AgentConfig, AgentFramework, AnyAgent
from any_agent.serving import A2AServingConfig
from any_agent.tools import a2a_tool, a2a_tool_async
from any_agent.tracing.agent_trace import AgentTrace
from tests.integration.helpers import DEFAULT_MODEL_ID, wait_for_server

from .conftest import (
    DATE_PROMPT,
    DEFAULT_TIMEOUT,
    a2a_client_from_agent,
    assert_contains_current_date_info,
    get_datetime,
)

if TYPE_CHECKING:
    from multiprocessing import Queue as QueueType


def _assert_valid_agent_trace(agent_trace: AgentTrace) -> None:
    """Assert that agent_trace is valid and has final output."""
    assert isinstance(agent_trace, AgentTrace)
    assert agent_trace.final_output


def _assert_has_date_agent_tool_call(agent_trace: AgentTrace) -> None:
    """Assert that the agent trace contains a tool execution span for the date agent."""
    assert any(
        span.is_tool_execution()
        and span.attributes.get("gen_ai.tool.name", None) == "call_date_agent"
        for span in agent_trace.spans
    )


@pytest.mark.asyncio
async def test_a2a_tool_async(agent_framework: AgentFramework) -> None:
    """Tests that an agent contacts another using A2A using the adapter tool.

    Note that there is an issue when using Google ADK: https://github.com/google/adk-python/pull/566
    """
    skip_reason = {
        AgentFramework.SMOLAGENTS: "async a2a is not supported",
    }
    if agent_framework in skip_reason:
        pytest.skip(
            f"Framework {agent_framework}, reason: {skip_reason[agent_framework]}"
        )

    env_check = validate_environment(DEFAULT_MODEL_ID)
    if not env_check["keys_in_environment"]:
        pytest.skip(f"{env_check['missing_keys']} needed for {agent_framework}")

    model_args: dict[str, Any] = (
        {"parallel_tool_calls": False}
        if agent_framework not in [AgentFramework.AGNO, AgentFramework.LLAMA_INDEX]
        else {}
    )
    model_args["temperature"] = 0.0

    # Create date agent
    date_agent_cfg = AgentConfig(
        instructions="Use the available tools to obtain additional information to answer the query.",
        name="date_agent",
        model_id=DEFAULT_MODEL_ID,
        description="Agent that can return the current date.",
        tools=[get_datetime],
        model_args=model_args,
    )
    date_agent = await AnyAgent.create_async(
        agent_framework=agent_framework,
        agent_config=date_agent_cfg,
    )

    # Serve the agent and get client
    tool_agent_endpoint = "tool_agent"
    serving_config = A2AServingConfig(
        port=0,
        endpoint=f"/{tool_agent_endpoint}",
        log_level="info",
    )

    async with a2a_client_from_agent(date_agent, serving_config) as (_, server_url):
        # Create main agent with A2A tool
        main_agent_cfg = AgentConfig(
            instructions="Use the available tools to obtain additional information to answer the query.",
            description="The orchestrator that can use other agents via tools using the A2A protocol.",
            model_id=DEFAULT_MODEL_ID,
            tools=[
                await a2a_tool_async(
                    server_url, http_kwargs={"timeout": DEFAULT_TIMEOUT}
                )
            ],
            model_args=model_args,
        )

        main_agent = await AnyAgent.create_async(
            agent_framework=agent_framework,
            agent_config=main_agent_cfg,
        )

        agent_trace = await main_agent.run_async(DATE_PROMPT)

        _assert_valid_agent_trace(agent_trace)
        assert_contains_current_date_info(str(agent_trace.final_output))
        _assert_has_date_agent_tool_call(agent_trace)


def _run_server(
    agent_framework_str: str,
    port: int,
    endpoint: str,
    model_id: str,
    server_queue: "QueueType[int]",
) -> None:
    """Run the server for the sync test."""
    date_agent_cfg = AgentConfig(
        instructions="Use the available tools to obtain additional information to answer the query.",
        name="date_agent",
        description="Agent that can return the current date.",
        tools=[get_datetime],
        model_id=model_id,
        model_args={"parallel_tool_calls": False}
        if agent_framework_str not in ["agno", "llama_index"]
        else None,
    )

    date_agent = AnyAgent.create(
        agent_framework=AgentFramework.from_string(agent_framework_str),
        agent_config=date_agent_cfg,
    )

    from any_agent.serving import A2AServingConfig, _get_a2a_app, serve_a2a

    serving_config = A2AServingConfig(
        port=port,
        endpoint=f"/{endpoint}",
        log_level="info",
    )

    app = _get_a2a_app(date_agent, serving_config=serving_config)

    serve_a2a(
        app,
        host=serving_config.host,
        port=serving_config.port,
        endpoint=serving_config.endpoint,
        log_level=serving_config.log_level,
        server_queue=server_queue,
    )


def test_a2a_tool_sync(agent_framework: AgentFramework) -> None:
    """Tests that an agent contacts another using A2A using the sync adapter tool.

    Note that there is an issue when using Google ADK: https://github.com/google/adk-python/pull/566
    """
    skip_reason = {
        AgentFramework.SMOLAGENTS: "async a2a is not supported; run_async_in_sync fails",
    }
    if agent_framework in skip_reason:
        pytest.skip(
            f"Framework {agent_framework}, reason: {skip_reason[agent_framework]}"
        )

    env_check = validate_environment(DEFAULT_MODEL_ID)
    if not env_check["keys_in_environment"]:
        pytest.skip(f"{env_check['missing_keys']} needed for {agent_framework}")

    server_process = None
    tool_agent_endpoint = "tool_agent_sync"

    server_queue = Queue()  # type: ignore[var-annotated]

    try:
        # Start the server in a separate process
        server_process = Process(
            target=_run_server,
            args=(
                agent_framework.value,
                0,
                tool_agent_endpoint,
                DEFAULT_MODEL_ID,
                server_queue,
            ),
        )
        server_process.start()

        test_port = server_queue.get()
        server_url = f"http://localhost:{test_port}/{tool_agent_endpoint}"
        wait_for_server(server_url)

        # Create main agent using sync methods
        main_agent_cfg = AgentConfig(
            instructions="Use the available tools to obtain additional information to answer the query.",
            description="The orchestrator that can use other agents via tools using the A2A protocol (sync version).",
            tools=[
                a2a_tool(
                    f"http://localhost:{test_port}/{tool_agent_endpoint}",
                    http_kwargs={"timeout": DEFAULT_TIMEOUT},
                )
            ],
            model_id=DEFAULT_MODEL_ID,
        )

        main_agent = AnyAgent.create(
            agent_framework=agent_framework,
            agent_config=main_agent_cfg,
        )

        agent_trace = main_agent.run(DATE_PROMPT)

        _assert_valid_agent_trace(agent_trace)
        assert_contains_current_date_info(str(agent_trace.final_output))
        _assert_has_date_agent_tool_call(agent_trace)

    finally:
        if server_process and server_process.is_alive():
            # Send SIGTERM for graceful shutdown
            server_process.terminate()
            server_process.join(timeout=10)
            if server_process.is_alive():
                # Force kill if graceful shutdown failed
                server_process.kill()
                server_process.join()



"""
This test is to test the task management functionality of the A2A classes.
Because the test isn't designed to test LLM performance, I mock the agent run_async method, this way
no calls are made to the LLM. I mock the responses to create a trace that mimics LLM output values,
and verify that the subsequent calls properly receive the previous conversation history.
"""

from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any
from uuid import uuid4

import httpx
import pytest
from a2a.client import A2AClient
from a2a.types import MessageSendParams, SendMessageRequest, TaskState
from pydantic import BaseModel

from any_agent import AgentConfig
from any_agent.config import AgentFramework
from any_agent.frameworks.any_agent import AnyAgent
from any_agent.frameworks.tinyagent import TinyAgent
from any_agent.serving import A2AServingConfig
from any_agent.serving.a2a.envelope import A2AEnvelope
from any_agent.tools.a2a import a2a_tool_async
from any_agent.tracing.agent_trace import AgentSpan, AgentTrace
from any_agent.tracing.otel_types import (
    Resource,
    SpanContext,
    SpanKind,
    Status,
)
from tests.integration.helpers import wait_for_server_async

if TYPE_CHECKING:
    from typing import Any


class TestResult(BaseModel):
    name: str
    job: str
    age: int | None = None


FIRST_TURN_PROMPT = "My name is Alice and I work as a software engineer."
FIRST_TURN_RESPONSE = "Hello, Alice! It's nice to meet you. You work as a software engineer. How old are you?"
SECOND_TURN_PROMPT = "What's my name, what do I do for work, and what's my age? Let me know if you need more information."
SECOND_TURN_RESPONSE = (
    "Your name is Alice, you work as a software engineer, and your age is 30."
)
THIRD_TURN_PROMPT = "My age is 30."
THIRD_TURN_RESPONSE = "Thank you for the information."


class MockConversationAgent(TinyAgent):
    """Mock agent implementation that simulates multi-turn conversation for testing."""

    def __init__(self, config: AgentConfig) -> None:
        super().__init__(config)
        self.output_type = A2AEnvelope[TestResult]
        self.turn_count = 0

    async def _load_agent(self) -> None:
        # Call parent's _load_agent to set up the basic structure
        await super()._load_agent()

    async def run_async(
        self, prompt: str, instrument: bool = True, **kwargs: Any
    ) -> AgentTrace:
        if self.turn_count == 0:
            # First turn: User introduces themselves
            assert FIRST_TURN_PROMPT in prompt
            self.turn_count += 1
            envelope = self.output_type(
                task_status=TaskState.completed,
                data=TestResult(
                    name="Alice",
                    job="software engineer",
                    age=None,
                ),
            )
            return self._create_mock_trace(
                envelope, FIRST_TURN_RESPONSE, FIRST_TURN_PROMPT
            )
        if self.turn_count == 1:
            # Second turn: User asks for information back
            assert FIRST_TURN_PROMPT in prompt
            assert SECOND_TURN_PROMPT in prompt
            self.turn_count += 1
            envelope = self.output_type(
                task_status=TaskState.input_required,
                data=TestResult(
                    name="Alice",
                    job="software engineer",
                    age=None,
                ),
            )
            return self._create_mock_trace(envelope, SECOND_TURN_RESPONSE, prompt)
        if self.turn_count == 2:
            # Third turn: User provides age
            assert FIRST_TURN_PROMPT in prompt
            assert SECOND_TURN_PROMPT in prompt
            assert THIRD_TURN_PROMPT in prompt
            self.turn_count += 1
            envelope = self.output_type(
                task_status=TaskState.completed,
                data=TestResult(
                    name="Alice",
                    job="software engineer",
                    age=30,
                ),
            )
            return self._create_mock_trace(envelope, THIRD_TURN_RESPONSE, prompt)
        msg = f"Unexpected turn count: {self.turn_count}"
        raise ValueError(msg)

    def _create_mock_trace(
        self, envelope: A2AEnvelope[TestResult], agent_response: str, prompt: str
    ) -> AgentTrace:
        """Create a mock AgentTrace with minimal spans for testing."""

        spans = []
        spans.append(
            AgentSpan(
                name="call_llm gpt-4o-mini",
                kind=SpanKind.INTERNAL,
                status=Status(),
                context=SpanContext(span_id=123),
                attributes={
                    "gen_ai.operation.name": "call_llm",
                    "gen_ai.request.model": "mock-model",
                    "gen_ai.input.messages": json.dumps(
                        [{"role": "user", "content": prompt}]
                    ),
                    "gen_ai.output": agent_response,
                    "gen_ai.output.type": "json",
                },
                links=[],
                events=[],
                resource=Resource(),
            )
        )

        return AgentTrace(
            spans=spans,
            final_output=envelope,
        )

    @classmethod
    def create(cls, framework: AgentFramework | str, config: AgentConfig) -> AnyAgent:
        return cls(config)


@pytest.mark.asyncio
async def test_a2a_tool_multiturn() -> None:
    """Test that agents can maintain conversation context across multiple interactions."""

    # Create a mock agent that simulates multi-turn conversation
    config = AgentConfig(
        model_id="gpt-4o-mini",  # Using real model ID but will be mocked
        instructions=(
            "You are a helpful assistant that remembers our conversation. "
            "When asked about previous information, reference what was said earlier. "
            "Keep your responses concise."
            " If you need more information, ask the user for it."
        ),
        description="Agent with conversation memory for testing session management.",
        output_type=TestResult,
    )

    agent = MockConversationAgent(config)

    # Configure session management with short timeout for testing
    serving_config = A2AServingConfig(
        port=0,
        task_timeout_minutes=2,  # Short timeout for testing
    )

    server_handle = await agent.serve_async(serving_config=serving_config)
    server_url = f"http://localhost:{server_handle.port}"
    await wait_for_server_async(server_url)

    try:
        async with httpx.AsyncClient(timeout=1500) as httpx_client:
            client = await A2AClient.get_client_from_agent_card_url(
                httpx_client, server_url
            )

            # First interaction - establish context
            first_message_id = str(uuid4())
            context_id = str(uuid4())  # This will be our session identifier

            send_message_payload_1 = {
                "message": {
                    "role": "user",
                    "parts": [
                        {
                            "kind": "text",
                            "text": FIRST_TURN_PROMPT,
                        }
                    ],
                    "messageId": first_message_id,
                    "contextId": context_id,  # Link messages to same conversation
                },
            }

            request_1 = SendMessageRequest(
                id=str(uuid4()), params=MessageSendParams(**send_message_payload_1)
            )
            response_1 = await client.send_message(request_1)

            assert response_1 is not None
            result = TestResult.model_validate_json(
                response_1.root.result.status.message.parts[0].root.text
            )
            assert result.name == "Alice"
            assert result.job.lower() == "software engineer"

            send_message_payload_2 = {
                "message": {
                    "role": "user",
                    "parts": [
                        {
                            "kind": "text",
                            "text": SECOND_TURN_PROMPT,
                        }
                    ],
                    "messageId": str(uuid4()),
                    "contextId": response_1.root.result.contextId,  # Same context to continue conversation
                    "taskId": response_1.root.result.id,
                },
            }

            request_2 = SendMessageRequest(
                id=str(uuid4()), params=MessageSendParams(**send_message_payload_2)
            )
            response_2 = await client.send_message(request_2)

            assert response_2 is not None
            result = TestResult.model_validate_json(
                response_2.root.result.status.message.parts[0].root.text
            )
            assert result.name == "Alice"
            assert result.job.lower() == "software engineer"
            assert result.age is None
            assert response_2.root.result.status.state == TaskState.input_required

            # Send a message to the agent to give the age
            send_message_payload_3 = {
                "message": {
                    "role": "user",
                    "parts": [{"kind": "text", "text": THIRD_TURN_PROMPT}],
                    "messageId": str(uuid4()),
                    "contextId": response_1.root.result.contextId,  # Same context to continue conversation
                    "taskId": response_1.root.result.id,
                },
            }
            request_3 = SendMessageRequest(
                id=str(uuid4()), params=MessageSendParams(**send_message_payload_3)
            )
            response_3 = await client.send_message(request_3)
            assert response_3 is not None
            result = TestResult.model_validate_json(
                response_3.root.result.status.message.parts[0].root.text
            )
            assert response_3.root.result.status.state == TaskState.completed
            assert result.age == 30

    finally:
        await server_handle.shutdown()


@pytest.mark.asyncio
async def test_a2a_tool_multiturn_async() -> None:
    """Test that agents can maintain conversation context across multiple interactions."""

    # Create a mock agent that simulates multi-turn conversation
    config = AgentConfig(
        model_id="gpt-4o-mini",  # Using real model ID but will be mocked
        instructions=(
            "You are a helpful assistant that remembers our conversation. "
            "When asked about previous information, reference what was said earlier. "
            "Keep your responses concise."
            " If you need more information, ask the user for it."
        ),
        name="Structured TestResult Agent",
        description="Agent with conversation memory for testing session management.",
        output_type=TestResult,
    )

    agent = MockConversationAgent(config)

    # Configure session management with short timeout for testing
    serving_config = A2AServingConfig(
        port=0,
        task_timeout_minutes=2,  # Short timeout for testing
    )

    server_handle = await agent.serve_async(serving_config=serving_config)
    server_url = f"http://localhost:{server_handle.port}"
    await wait_for_server_async(server_url)
    try:

        class MainAgentAnswer(BaseModel):
            first_turn_success: bool
            second_turn_success: bool
            third_turn_success: bool

        main_agent_cfg = AgentConfig(
            model_id="gpt-4.1-nano",
            instructions="Use the available tools to obtain additional information to answer the query.",
            description="The orchestrator that can use other agents via tools using the A2A protocol.",
            tools=[await a2a_tool_async(server_url, http_kwargs={"timeout": 10.0})],
            output_type=MainAgentAnswer,
            model_args={
                "parallel_tool_calls": False  # to force it to talk to the agent one call at a time
            },
        )

        main_agent = await AnyAgent.create_async(
            agent_framework=AgentFramework.TINYAGENT,
            agent_config=main_agent_cfg,
        )
        prompt = f"""
        Please talk to the structured testresult agent and interact with it. You'll contact it to ask three questions. Say the exact words from the prompt in your query to the agent.

        1. {FIRST_TURN_PROMPT}
        2. {SECOND_TURN_PROMPT}
        3. {THIRD_TURN_PROMPT}

        Make sure you appropriately continue the conversation by providing it with the task id if you want to continue the conversation.
        """

        agent_trace = await main_agent.run_async(prompt)
        assert agent_trace.final_output is not None
        assert isinstance(agent_trace.final_output, MainAgentAnswer)
        assert agent_trace.final_output.first_turn_success
        assert agent_trace.final_output.second_turn_success
        assert agent_trace.final_output.third_turn_success
    finally:
        await server_handle.shutdown()



# mypy: disable-error-code="no-untyped-def,union-attr"
from __future__ import annotations

import json
from typing import TYPE_CHECKING

from opentelemetry.trace import StatusCode

from .common import _set_tool_output

if TYPE_CHECKING:
    from agents import GenerationSpanData
    from agents.tracing import TracingProcessor
    from opentelemetry.trace import Span

    from any_agent.frameworks.openai import OpenAIAgent


def _set_llm_input(span_data: GenerationSpanData, span: Span) -> None:
    if input_messages := span_data.input:
        span.set_attribute(
            "gen_ai.input.messages",
            json.dumps(
                input_messages,
                default=str,
                ensure_ascii=False,
            ),
        )


def _set_llm_output(span_data: GenerationSpanData, span: Span) -> None:
    if not span_data.output:
        return
    output = span_data.output[0]
    if content := output.get("content"):
        span.set_attributes(
            {
                "gen_ai.output": content,
                "gen_ai.output.type": "text",
            }
        )
    if tool_calls := output.get("tool_calls"):
        span.set_attributes(
            {
                "gen_ai.output": json.dumps(
                    [
                        {
                            "tool.name": tool_call.get("function", {}).get(
                                "name", "No name"
                            ),
                            "tool.args": tool_call.get("function", {}).get(
                                "arguments", "No args"
                            ),
                        }
                        for tool_call in tool_calls
                    ]
                ),
                "gen_ai.output.type": "json",
            }
        )
    if token_usage := span_data.usage:
        span.set_attributes(
            {
                "gen_ai.usage.input_tokens": token_usage["input_tokens"],
                "gen_ai.usage.output_tokens": token_usage["output_tokens"],
            }
        )


class _OpenAIAgentsInstrumentor:
    def __init__(self) -> None:
        self.first_llm_calls: set[int] = set()
        self.current_spans: dict[str, Span] = {}
        self._processor: TracingProcessor | None = None

    def instrument(self, agent: OpenAIAgent) -> None:
        if len(agent._running_traces) > 1:
            return
        from agents import FunctionSpanData, GenerationSpanData
        from agents.tracing import TracingProcessor
        from agents.tracing.processors import BatchTraceProcessor

        first_llm_calls = self.first_llm_calls
        current_spans = self.current_spans

        tracer = agent._tracer

        class AnyAgentTracingProcessor(TracingProcessor):
            def on_trace_start(self, trace):
                pass

            def on_trace_end(self, trace):
                pass

            def on_span_start(self, span):
                span_data = span.span_data
                if isinstance(span_data, GenerationSpanData):
                    model = str(span_data.model)
                    otel_span = tracer.start_span(
                        name=f"call_llm {model}",
                    )
                    otel_span.set_attributes(
                        {
                            "gen_ai.operation.name": "call_llm",
                            "gen_ai.request.model": model,
                        }
                    )
                    current_spans[span.span_id] = otel_span
                elif isinstance(span_data, FunctionSpanData):
                    otel_span = tracer.start_span(
                        name=f"execute_tool {span_data.name}",
                    )
                    otel_span.set_attributes(
                        {
                            "gen_ai.operation.name": "execute_tool",
                            "gen_ai.tool.name": span_data.name,
                        }
                    )
                    current_spans[span.span_id] = otel_span

            def on_span_end(self, span):
                if span.span_id not in current_spans:
                    return
                span_data = span.span_data
                if isinstance(span_data, GenerationSpanData):
                    otel_span = current_spans[span.span_id]
                    trace_id = otel_span.get_span_context().trace_id
                    if trace_id not in first_llm_calls:
                        first_llm_calls.add(trace_id)
                        _set_llm_input(span_data, otel_span)
                    _set_llm_output(span_data, otel_span)
                    otel_span.set_status(StatusCode.OK)
                    otel_span.end()
                    if trace_id in agent._running_traces:
                        agent._running_traces[trace_id].add_span(otel_span)
                    del current_spans[span.span_id]
                elif isinstance(span_data, FunctionSpanData):
                    otel_span = current_spans[span.span_id]
                    otel_span.set_attributes(
                        {
                            "gen_ai.tool.args": span_data.input or "{}",
                        }
                    )
                    _set_tool_output(span_data.output, otel_span)
                    otel_span.end()
                    trace_id = otel_span.get_span_context().trace_id
                    if trace_id in agent._running_traces:
                        agent._running_traces[trace_id].add_span(otel_span)
                    del current_spans[span.span_id]

            def force_flush(self):
                pass

            def shutdown(self):
                pass

        self._processor = AnyAgentTracingProcessor()
        from agents.tracing.setup import GLOBAL_TRACE_PROVIDER

        with GLOBAL_TRACE_PROVIDER._multi_processor._lock:
            GLOBAL_TRACE_PROVIDER._multi_processor._processors = tuple(
                p
                for p in GLOBAL_TRACE_PROVIDER._multi_processor._processors
                if not isinstance(p, BatchTraceProcessor)
            )
            GLOBAL_TRACE_PROVIDER._multi_processor._processors += (self._processor,)

    def uninstrument(self, agent: OpenAIAgent) -> None:
        if len(agent._running_traces) > 1:
            return
        from agents.tracing.setup import GLOBAL_TRACE_PROVIDER

        if self._processor:
            with GLOBAL_TRACE_PROVIDER._multi_processor._lock:
                GLOBAL_TRACE_PROVIDER._multi_processor._processors = tuple(
                    p
                    for p in GLOBAL_TRACE_PROVIDER._multi_processor._processors
                    if p is not self._processor
                )



# pylint: disable=unused-argument, unused-variable
from collections.abc import Sequence
from typing import Any

import pytest

from any_agent.config import AgentFramework, MCPParams, MCPSse, Tool
from any_agent.tools import _get_mcp_server, _MCPConnection


@pytest.mark.asyncio
async def test_sse_tool_filtering(
    agent_framework: AgentFramework,
    sse_params_echo_server: MCPSse,
    tools: Sequence[str],
) -> None:
    reduced_tools = tools[:-1]
    new_params = sse_params_echo_server.model_copy(update={"tools": reduced_tools})
    server = _get_mcp_server(new_params, agent_framework)
    await server._setup_tools()
    if agent_framework is AgentFramework.AGNO:
        # Check that only the specified tools are included
        assert set(server.tools[0].functions.keys()) == set(reduced_tools)  # type: ignore[union-attr]
    else:
        assert len(server.tools) == len(reduced_tools)  # ignore[arg-type]


@pytest.mark.asyncio
async def test_mcp_tools_loaded(
    agent_framework: AgentFramework,
    mcp_params: MCPParams,
    mcp_connection: _MCPConnection[Any],
    tools: Sequence[Tool],
) -> None:
    mcp_server = _get_mcp_server(mcp_params, agent_framework)

    await mcp_server._setup_tools(mcp_connection=mcp_connection)

    assert mcp_server.tools == list(tools)


def test_filter_tools(
    mcp_connection: _MCPConnection[Any], mcp_params: MCPParams
) -> None:
    """Test the _filter_tools method of _MCPConnection."""

    # Create tools with names
    class NamedTool:
        def __init__(self, name: str):
            self.name = name

    all_tools = [NamedTool("tool1"), NamedTool("tool2"), NamedTool("tool3")]

    # Case 1: No tools requested, all tools returned
    mcp_connection.mcp_tool = mcp_params.model_copy(update={"tools": None})
    filtered_tools = mcp_connection._filter_tools(all_tools)
    assert filtered_tools == all_tools

    # Case 2: Some tools requested, only those tools returned in the right order
    mcp_connection.mcp_tool = mcp_params.model_copy(
        update={"tools": ["tool3", "tool1"]}
    )
    filtered_tools = mcp_connection._filter_tools(all_tools)
    assert len(filtered_tools) == 2
    assert filtered_tools[0].name == "tool3"
    assert filtered_tools[1].name == "tool1"

    # Case 3: Request for non-existent tool raises ValueError
    mcp_connection.mcp_tool = mcp_params.model_copy(
        update={"tools": ["tool1", "nonexistent"]}
    )
    with pytest.raises(ValueError) as excinfo:  # noqa: PT011
        mcp_connection._filter_tools(all_tools)
    assert "Missing: ['nonexistent']" in str(excinfo.value)



import asyncio
import time

import httpx
import requests

DEFAULT_MODEL_ID = "gpt-4.1-nano"


def wait_for_server(
    server_url: str, max_attempts: int = 20, poll_interval: float = 0.5
) -> None:
    attempts = 0
    while True:
        try:
            # Try to make a basic GET request to check if server is responding
            requests.get(server_url, timeout=1.0)
            return  # noqa: TRY300
        except (requests.RequestException, ConnectionError):
            # Server not ready yet, continue polling
            pass

        time.sleep(poll_interval)
        attempts += 1
        if attempts >= max_attempts:
            msg = f"Could not connect to {server_url}. Tried {max_attempts} times with {poll_interval} second interval."
            raise ConnectionError(msg)


async def wait_for_server_async(
    server_url: str, max_attempts: int = 20, poll_interval: float = 0.5
) -> None:
    attempts = 0

    async with httpx.AsyncClient() as client:
        while True:
            try:
                # Try to make a basic GET request to check if server is responding
                await client.get(server_url, timeout=1.0)
                return  # noqa: TRY300
            except (httpx.RequestError, httpx.TimeoutException):
                # Server not ready yet, continue polling
                pass

            await asyncio.sleep(poll_interval)
            attempts += 1
            if attempts >= max_attempts:
                msg = f"Could not connect to {server_url}. Tried {max_attempts} times with {poll_interval} second interval."
                raise ConnectionError(msg)



from rich.prompt import Prompt

from any_agent.logging import logger


def show_plan(plan: str) -> str:
    """Show the current plan to the user.

    Args:
        plan: The current plan.

    """
    logger.info(f"Current plan: {plan}")
    return plan


def show_final_output(answer: str) -> str:
    """Show the final answer to the user.

    Args:
        answer: The final answer.

    """
    logger.info(f"Final output: {answer}")
    return answer


def ask_user_verification(query: str) -> str:
    """Asks user to verify the given `query`.

    Args:
        query: The question that requires verification.

    """
    return input(f"{query} => Type your answer here:")


def send_console_message(user: str, query: str) -> str:
    """Send the specified user a message via console and returns their response.

    Args:
        query: The question to ask the user.
        user: The user to ask the question to.

    Returns:
        str: The user's response.

    """
    return Prompt.ask(f"{query}\n{user}")



# mypy: disable-error-code="method-assign,no-untyped-call,no-untyped-def,union-attr"
from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any

from opentelemetry.trace import Span, StatusCode

from any_agent.tracing.instrumentation.common import _set_tool_output

if TYPE_CHECKING:
    from collections.abc import Sequence

    from llama_index.core.agent.workflow.workflow_events import AgentOutput
    from llama_index.core.base.llms.types import ChatMessage
    from llama_index.core.llms.llm import ToolSelection
    from llama_index.core.memory import BaseMemory
    from llama_index.core.tools.types import AsyncBaseTool
    from llama_index.core.workflow import Context

    from any_agent.frameworks.llama_index import LlamaIndexAgent


def _set_llm_input(messages: list[ChatMessage], span: Span) -> None:
    span.set_attribute(
        "gen_ai.input.messages",
        json.dumps(
            [
                {
                    "role": message.role.value,
                    "content": message.content or "No content",
                }
                for message in messages
            ]
        ),
    )


def _set_llm_output(output: AgentOutput, span: Span) -> None:
    if response := output.response:
        if content := response.content:
            span.set_attributes(
                {
                    "gen_ai.output": content,
                    "gen_ai.output.type": "text",
                }
            )
    tool_calls: list[ToolSelection]
    if tool_calls := output.tool_calls:
        span.set_attributes(
            {
                "gen_ai.output": json.dumps(
                    [
                        {
                            "tool.name": getattr(tool_call, "tool_name", "No name"),
                            "tool.args": getattr(tool_call, "tool_kwars", "{}"),
                        }
                        for tool_call in tool_calls
                    ],
                    default=str,
                    ensure_ascii=False,
                ),
                "gen_ai.output.type": "json",
            }
        )
    raw: dict[str, Any] | None
    if raw := getattr(output, "raw", None):
        token_usage: dict[str, int]
        if token_usage := raw.get("usage"):
            span.set_attributes(
                {
                    "gen_ai.usage.input_tokens": token_usage.get("prompt_tokens", 0),
                    "gen_ai.usage.output_tokens": token_usage.get(
                        "completion_tokens", 0
                    ),
                }
            )


class _LlamaIndexInstrumentor:
    def __init__(self) -> None:
        self.first_llm_calls: set[int] = set()
        self._original_take_step: Any | None = None
        self._original_acalls: dict[str, Any] = {}

    def instrument(self, agent: LlamaIndexAgent) -> None:
        if len(agent._running_traces) > 1:
            return

        tracer = agent._tracer

        self._original_take_step = agent._agent.take_step

        async def wrap_take_step(
            ctx: Context[Any],
            llm_input: list[ChatMessage],
            tools: Sequence[AsyncBaseTool],
            memory: BaseMemory,
        ) -> AgentOutput:
            model = getattr(agent._agent.llm, "model", "No model")
            with tracer.start_as_current_span(f"call_llm {model}") as span:
                span.set_attributes(
                    {
                        "gen_ai.operation.name": "call_llm",
                        "gen_ai.request.model": model,
                    }
                )
                trace_id = span.get_span_context().trace_id
                if trace_id not in self.first_llm_calls:
                    self.first_llm_calls.add(trace_id)
                    _set_llm_input(llm_input, span)

                output: AgentOutput = await self._original_take_step(  # type: ignore[misc]
                    ctx, llm_input, tools, memory
                )
                if output:
                    _set_llm_output(output, span)

                span.set_status(StatusCode.OK)
                agent._running_traces[trace_id].add_span(span)

            return output

        class WrappedAcall:
            def __init__(self, metadata, original_acall):
                self.metadata = metadata
                self.original_acall = original_acall

            async def acall(self, *args, ctx=None, **kwargs):
                name = self.metadata.name
                with tracer.start_as_current_span(f"execute_tool {name}") as span:
                    span.set_attributes(
                        {
                            "gen_ai.operation.name": "execute_tool",
                            "gen_ai.tool.name": name,
                            "gen_ai.tool.args": json.dumps(
                                kwargs, default=str, ensure_ascii=False
                            ),
                        }
                    )

                    output = await self.original_acall(**kwargs)
                    if raw_output := getattr(output, "raw_output", None):
                        if content := getattr(raw_output, "content", None):
                            _set_tool_output(content[0].text, span)
                        else:
                            _set_tool_output(raw_output, span)
                    else:
                        _set_tool_output(output, span)

                    trace_id = span.get_span_context().trace_id
                    agent._running_traces[trace_id].add_span(span)
                    return output

        for tool in agent._agent.tools:
            self._original_acalls[str(tool.metadata.name)] = tool.acall
            wrapped = WrappedAcall(tool.metadata, tool.acall)
            tool.acall = wrapped.acall

        # bypass Pydantic validation because _agent is a BaseModel
        agent._agent.model_config["extra"] = "allow"
        agent._agent.take_step = wrap_take_step

    def uninstrument(self, agent: LlamaIndexAgent) -> None:
        if len(agent._running_traces) > 1:
            return

        if self._original_take_step:
            agent._agent.take_step = self._original_take_step
        if self._original_acalls:
            for tool in agent._agent.tools:
                tool.acall = self._original_acalls[str(tool.metadata.name)]



import json
import os
import subprocess
import time
from datetime import datetime, timedelta
from pathlib import Path
from typing import Any

import pytest
from litellm.utils import validate_environment
from pydantic import BaseModel, ConfigDict

from any_agent import (
    AgentConfig,
    AgentFramework,
    AnyAgent,
)
from any_agent.config import MCPStdio
from any_agent.evaluation.agent_judge import AgentJudge
from any_agent.evaluation.llm_judge import LlmJudge
from any_agent.evaluation.schemas import EvaluationOutput
from any_agent.tracing import TRACE_PROVIDER
from any_agent.tracing.agent_trace import AgentSpan, AgentTrace, CostInfo, TokenInfo
from any_agent.tracing.exporter import _ConsoleExporter


def uvx_installed() -> bool:
    try:
        result = subprocess.run(
            ["uvx", "--version"],  # noqa: S607
            capture_output=True,
            check=True,
        )
        return True if result.returncode == 0 else False  # noqa: TRY300
    except Exception:
        return False


def assert_trace(agent_trace: AgentTrace, agent_framework: AgentFramework) -> None:
    def assert_first_llm_call(llm_call: AgentSpan) -> None:
        """Checks the `_set_llm_inputs` implemented by each framework's instrumentation."""
        assert llm_call.attributes.get("gen_ai.input.messages", None) is not None
        # input.messages should be a valid JSON string (list of dicts)
        input_messages_raw = llm_call.attributes.get("gen_ai.input.messages")
        assert input_messages_raw is not None
        input_messages = json.loads(input_messages_raw)
        assert input_messages[0]["role"] == "system"
        assert input_messages[1]["role"] == "user"

    def assert_first_tool_execution(tool_execution: AgentSpan) -> None:
        """Checks the tools setup implemented by each framework's instrumentation."""
        assert tool_execution.attributes.get("gen_ai.tool.args", None) is not None
        # tool.args should be a JSON string (dict)
        tool_args_raw = tool_execution.attributes.get("gen_ai.tool.args")
        assert tool_args_raw is not None
        args = json.loads(tool_args_raw)
        assert "timezone" in args
        assert isinstance(agent_trace, AgentTrace)
        assert agent_trace.final_output

    agent_invocations = []
    llm_calls = []
    tool_executions = []
    for span in agent_trace.spans:
        if span.is_agent_invocation():
            agent_invocations.append(span)
        elif span.is_llm_call():
            llm_calls.append(span)
        elif span.is_tool_execution():
            tool_executions.append(span)
        else:
            msg = f"Unexpected span: {span}"
            raise AssertionError(msg)

    assert len(agent_invocations) == 1

    assert len(llm_calls) >= 2
    assert_first_llm_call(llm_calls[0])

    assert len(tool_executions) >= 2
    assert_first_tool_execution(tool_executions[0])

    messages = agent_trace.spans_to_messages()
    assert messages[0].role == "system"
    assert messages[1].role == "user"
    assert len(messages) == 2 + len(llm_calls) + len(tool_executions)


def assert_duration(agent_trace: AgentTrace, wall_time_s: float) -> None:
    assert agent_trace.duration is not None
    assert isinstance(agent_trace.duration, timedelta)
    assert agent_trace.duration.total_seconds() > 0

    diff = abs(agent_trace.duration.total_seconds() - wall_time_s)
    assert diff < 0.1, (
        f"duration ({agent_trace.duration.total_seconds()}s) and wall_time ({wall_time_s}s) differ by more than 0.1s: {diff}s"
    )


def assert_cost(agent_trace: AgentTrace) -> None:
    assert isinstance(agent_trace.cost, CostInfo)
    assert agent_trace.cost.input_cost > 0
    assert agent_trace.cost.output_cost > 0
    assert agent_trace.cost.input_cost + agent_trace.cost.output_cost < 1.00


def assert_tokens(agent_trace: AgentTrace) -> None:
    assert isinstance(agent_trace.tokens, TokenInfo)
    assert agent_trace.tokens.input_tokens > 0
    assert agent_trace.tokens.output_tokens > 0
    assert (agent_trace.tokens.input_tokens + agent_trace.tokens.output_tokens) < 20000


def assert_eval(agent_trace: AgentTrace) -> None:
    """Test evaluation using the new judge classes."""
    # Test 1: Check if agent called write_file tool using LlmJudge
    llm_judge = LlmJudge(model_id="gpt-4.1-nano")
    result1 = llm_judge.run(
        context=str(agent_trace.spans_to_messages()),
        question="Did the agent call the write_file tool during execution?",
    )
    assert isinstance(result1, EvaluationOutput)
    assert result1.passed, (
        f"Expected agent to call write_file tool, but evaluation failed: {result1.reasoning}"
    )

    # Test 2: Check if agent wrote the current year to file using AgentJudge
    agent_judge = AgentJudge(model_id="gpt-4.1-mini")

    def get_current_year() -> str:
        """Get the current year"""
        return str(datetime.now().year)

    result2 = agent_judge.run(
        trace=agent_trace,
        question="Did the agent write the current year to a file?",
        additional_tools=[get_current_year],
    )
    assert isinstance(result2, EvaluationOutput)
    assert result2.passed, (
        f"Expected agent to write current year to file, but evaluation failed: {result2.reasoning}"
    )

    # Test 3: Verify at least one evaluation passes (basic sanity check)
    results = [result1, result2]
    passed_count = sum(1 for r in results if r.passed)
    assert passed_count >= 1, (
        f"Expected at least 1 evaluation to pass, but got {passed_count}/2"
    )


class Step(BaseModel):
    number: int
    description: str


class Steps(BaseModel):
    model_config = ConfigDict(extra="forbid")
    steps: list[Step]


def test_load_and_run_agent(
    agent_framework: AgentFramework, tmp_path: Path, request: pytest.FixtureRequest
) -> None:
    kwargs = {}

    tmp_file = "tmp.txt"

    if not uvx_installed():
        msg = "uvx is not installed. Please install it to run this test."
        raise RuntimeError(msg)

    def write_file(text: str) -> None:
        """write the text to a file in the tmp_path directory

        Args:
            text (str): The text to write to the file.

        Returns:
            None
        """
        with open(os.path.join(tmp_path, tmp_file), "w", encoding="utf-8") as f:
            f.write(text)

    kwargs["model_id"] = "gpt-4.1-mini"
    env_check = validate_environment(kwargs["model_id"])
    if not env_check["keys_in_environment"]:
        pytest.skip(f"{env_check['missing_keys']} needed for {agent_framework}")

    model_args: dict[str, Any] = (
        {"parallel_tool_calls": False}
        if agent_framework not in [AgentFramework.AGNO, AgentFramework.LLAMA_INDEX]
        else {}
    )
    model_args["temperature"] = 0.0
    tools = [
        write_file,
        MCPStdio(
            command="uvx",
            args=["mcp-server-time", "--local-timezone=America/New_York"],
            tools=[
                "get_current_time",
            ],
        ),
    ]
    agent_config = AgentConfig(
        tools=tools,  # type: ignore[arg-type]
        instructions="Use the available tools to answer.",
        model_args=model_args,
        output_type=Steps,
        **kwargs,  # type: ignore[arg-type]
    )
    agent = AnyAgent.create(agent_framework, agent_config)
    update_trace = request.config.getoption("--update-trace-assets")
    if update_trace:
        with TRACE_PROVIDER._active_span_processor._lock:  # type: ignore[attr-defined]
            for p in TRACE_PROVIDER._active_span_processor._span_processors:  # type: ignore[attr-defined]
                if isinstance(p.span_exporter, _ConsoleExporter):
                    console = p.span_exporter.console
                    console.record = True

    start_ns = time.time_ns()
    agent_trace = agent.run(
        "Find what year it is in the America/New_York timezone and write the value (single number) to a file. "
        "Finally, return a list of the steps you have taken.",
    )
    end_ns = time.time_ns()

    assert isinstance(agent_trace.final_output, Steps)

    assert (tmp_path / tmp_file).read_text() == str(datetime.now().year)

    assert_trace(agent_trace, agent_framework)
    assert_duration(agent_trace, (end_ns - start_ns) / 1_000_000_000)
    assert_cost(agent_trace)
    assert_tokens(agent_trace)

    if update_trace:
        trace_path = Path(__file__).parent.parent / "assets" / agent_framework.name
        with open(f"{trace_path}_trace.json", "w", encoding="utf-8") as f:
            f.write(agent_trace.model_dump_json(indent=2))
            f.write("\n")
        html_output = console.export_html(inline_styles=True)
        with open(f"{trace_path}_trace.html", "w", encoding="utf-8") as f:
            f.write(html_output.replace("<!DOCTYPE html>", ""))

    assert_eval(agent_trace)



"""Utilities for running async code in sync contexts."""

from __future__ import annotations

import asyncio
import concurrent.futures
from typing import TYPE_CHECKING, Any, TypeVar

T = TypeVar("T")

if TYPE_CHECKING:
    from collections.abc import Coroutine


def run_async_in_sync(coro: Coroutine[Any, Any, T]) -> T:
    """Run an async coroutine in a synchronous context.

    Handles different event loop scenarios:
    - If a loop is running, uses threading to avoid conflicts
    - If no loop exists, creates one or uses the current loop

    Args:
        coro: The coroutine to execute

    Returns:
        The result of the coroutine execution

    """
    try:
        # Check if there's a running event loop
        asyncio.get_running_loop()

        # If we get here, there's a loop running, so we can't use run_until_complete()
        # or asyncio.run() - must use threading approach
        def run_in_thread() -> T:
            return asyncio.run(coro)

        with concurrent.futures.ThreadPoolExecutor() as executor:
            return executor.submit(run_in_thread).result()

    except RuntimeError:
        # No running event loop - try to get available loop
        try:
            loop = asyncio.get_event_loop()
            return loop.run_until_complete(coro)
        except RuntimeError:
            # No event loop at all - create one
            return asyncio.run(coro)



# mypy: disable-error-code="method-assign,no-untyped-def"
from __future__ import annotations

import json
from copy import deepcopy
from typing import TYPE_CHECKING, Any

from opentelemetry.trace import StatusCode

from .common import _set_tool_output

if TYPE_CHECKING:
    from collections.abc import Callable

    from litellm.types.utils import (
        ChatCompletionMessageToolCall,
        ModelResponse,
        Usage,
    )
    from opentelemetry.trace import Span

    from any_agent.frameworks.tinyagent import TinyAgent


def _set_llm_input(messages: list[dict[str, str]], span: Span) -> None:
    span.set_attribute(
        "gen_ai.input.messages", json.dumps(messages, default=str, ensure_ascii=False)
    )


def _set_llm_output(response: ModelResponse, span: Span) -> None:
    if not response.choices:
        return

    message = getattr(response.choices[0], "message", None)
    if not message:
        return

    if content := getattr(message, "content", None):
        span.set_attributes(
            {
                "gen_ai.output": content,
                "gen_ai.output.type": "text",
            }
        )
    tool_calls: list[ChatCompletionMessageToolCall] | None
    if tool_calls := getattr(message, "tool_calls", None):
        span.set_attributes(
            {
                "gen_ai.output": json.dumps(
                    [
                        {
                            "tool.name": getattr(tool_call.function, "name", "No name"),
                            "tool.args": getattr(
                                tool_call.function, "arguments", "No name"
                            ),
                        }
                        for tool_call in tool_calls
                        if tool_call.function
                    ],
                    default=str,
                    ensure_ascii=False,
                ),
                "gen_ai.output.type": "json",
            }
        )

    token_usage: Usage | None
    if token_usage := getattr(response, "model_extra", {}).get("usage"):
        if token_usage:
            span.set_attributes(
                {
                    "gen_ai.usage.input_tokens": token_usage.prompt_tokens,
                    "gen_ai.usage.output_tokens": token_usage.completion_tokens,
                }
            )


class _TinyAgentInstrumentor:
    def __init__(self) -> None:
        self.first_llm_calls: set[int] = set()
        self._original_call_model: Callable[..., Any] | None = None
        self._original_call_tool: Callable[..., Any] | None = None

    def instrument(self, agent: TinyAgent) -> None:
        if len(agent._running_traces) > 1:
            return

        tracer = agent._tracer
        self._original_call_model = agent.call_model

        async def call_model(**kwargs):
            model = kwargs.get("model", "No model")
            with tracer.start_as_current_span(f"call_llm {model}") as span:
                span.set_attributes(
                    {
                        "gen_ai.operation.name": "call_llm",
                        "gen_ai.request.model": model,
                    }
                )
                trace_id = span.get_span_context().trace_id
                if trace_id not in self.first_llm_calls:
                    self.first_llm_calls.add(trace_id)
                    _set_llm_input(kwargs["messages"], span)

                response: ModelResponse = await self._original_call_model(**kwargs)  # type: ignore[misc]

                if response_model := getattr(response, "model", None):
                    span.set_attribute("gen_ai.response.model", response_model)

                _set_llm_output(response, span)

                span.set_status(StatusCode.OK)
                agent._running_traces[trace_id].add_span(span)

                return response

        agent.call_model = call_model

        class WrappedCallTool:
            def __init__(self, original_call_tool):
                self.original_call_tool = original_call_tool

            async def call_tool(self, request: dict[str, Any]):
                with tracer.start_as_current_span(
                    f"execute_tool {request.get('name')}"
                ) as span:
                    span.set_attributes(
                        {
                            "gen_ai.operation.name": "execute_tool",
                            "gen_ai.tool.name": request.get("name", "No name"),
                            "gen_ai.tool.args": json.dumps(
                                request.get("arguments", {}),
                                default=str,
                                ensure_ascii=False,
                            ),
                        }
                    )

                    result = await self.original_call_tool(request)

                    _set_tool_output(result, span)

                    trace_id = span.get_span_context().trace_id
                    agent._running_traces[trace_id].add_span(span)

                    return result

        self._original_clients = deepcopy(agent.clients)
        wrapped_tools = {}
        for key, tool in agent.clients.items():
            wrapped = WrappedCallTool(tool.call_tool)  # type: ignore[no-untyped-call]
            tool.call_tool = wrapped.call_tool
            wrapped_tools[key] = tool
        agent.clients = wrapped_tools

    def uninstrument(self, agent: TinyAgent) -> None:
        if len(agent._running_traces) > 1:
            return
        if self._original_call_model:
            agent.call_model = self._original_call_model
        if self._original_clients:
            agent.clients = self._original_clients



import multiprocessing
from uuid import uuid4

import httpx
import pytest
from a2a.client import A2AClient

# Import your agent and config
from any_agent import AgentConfig, AnyAgent
from any_agent.serving import A2AServingConfig
from tests.integration.helpers import DEFAULT_MODEL_ID, wait_for_server_async

from .conftest import A2ATestHelpers, a2a_client_from_agent


def serve_agent(port: int) -> None:
    agent = AnyAgent.create(
        "tinyagent",
        AgentConfig(
            model_id=DEFAULT_MODEL_ID,
            instructions="Directly answer the question without asking the user for input.",
            description="I'm an agent to help.",
        ),
    )
    agent.serve(serving_config=A2AServingConfig(port=port))


@pytest.mark.asyncio
async def test_serve_sync(test_port: int, a2a_test_helpers: A2ATestHelpers) -> None:
    # Start the agent in a subprocess
    proc = multiprocessing.Process(target=serve_agent, args=(test_port,), daemon=True)
    proc.start()
    server_url = f"http://localhost:{test_port}"
    await wait_for_server_async(server_url)

    try:
        async with httpx.AsyncClient(timeout=10.0) as httpx_client:
            client = await A2AClient.get_client_from_agent_card_url(
                httpx_client, server_url
            )

            request = a2a_test_helpers.create_send_message_request(
                text="What is an agent?"
            )
            response = await client.send_message(request)
            assert response is not None
    finally:
        proc.kill()
        proc.join()


@pytest.mark.asyncio
async def test_serve_async(test_port: int, a2a_test_helpers: A2ATestHelpers) -> None:
    # Create and serve the agent
    agent = await AnyAgent.create_async(
        "tinyagent",
        AgentConfig(
            model_id=DEFAULT_MODEL_ID,
            instructions="Directly answer the question without asking the user for input.",
            description="I'm an agent to help.",
        ),
    )

    # Use the context manager for proper cleanup
    async with a2a_client_from_agent(agent, A2AServingConfig(port=test_port)) as (
        client,
        _,
    ):
        request = a2a_test_helpers.create_send_message_request(
            text="What is an agent?",
            message_id=uuid4().hex,
        )
        response = await client.send_message(request)
        assert response is not None



# mypy: disable-error-code="arg-type,union-attr"
from __future__ import annotations

import json
from typing import TYPE_CHECKING

from opentelemetry.sdk.trace.export import (
    SpanExporter,
    SpanExportResult,
)
from rich.console import Console, Group
from rich.json import JSON
from rich.markdown import Markdown
from rich.panel import Panel

if TYPE_CHECKING:
    from collections.abc import Sequence

    from opentelemetry.sdk.trace import ReadableSpan


SCOPE_NAME = "any_agent"


def _get_output_panel(span: ReadableSpan) -> Panel | None:
    if output := span.attributes.get("gen_ai.output", None):
        output_type = span.attributes.get("gen_ai.output.type", "text")
        return Panel(
            Markdown(output) if output_type != "json" else JSON(output),
            title="OUTPUT",
            style="white",
            title_align="left",
        )
    return None


class _ConsoleExporter(SpanExporter):
    def __init__(self) -> None:
        self.console: Console = Console()

    def print_to_console(self, span: ReadableSpan) -> None:
        if scope := span.instrumentation_scope:
            if scope.name != SCOPE_NAME:
                return

        if not span.attributes:
            return

        operation_name = span.attributes.get("gen_ai.operation.name", "")

        if operation_name == "call_llm":
            panels = []
            if messages := span.attributes.get("gen_ai.input.messages"):
                panels.append(
                    Panel(
                        JSON(messages), title="INPUT", style="white", title_align="left"
                    )
                )
            if output_panel := _get_output_panel(span):
                panels.append(output_panel)
            if usage := {
                k.replace("gen_ai.usage.", ""): v
                for k, v in span.attributes.items()
                if "usage" in k
            }:
                panels.append(
                    Panel(
                        JSON(json.dumps(usage)),
                        title="USAGE",
                        style="white",
                        title_align="left",
                    )
                )
            self.console.print(
                Panel(
                    Group(*panels),
                    title=f"{operation_name.upper()}: {span.attributes.get('gen_ai.request.model')}",
                    style="yellow",
                )
            )
        elif operation_name == "execute_tool":
            panels = [
                Panel(
                    JSON(span.attributes.get("gen_ai.tool.args", "{}")),
                    title="Input",
                    style="white",
                    title_align="left",
                )
            ]
            if output_panel := _get_output_panel(span):
                panels.append(output_panel)
            self.console.print(
                Panel(
                    Group(*panels),
                    title=f"{operation_name.upper()}: {span.attributes.get('gen_ai.tool.name')}",
                    style="blue",
                )
            )

    def export(self, spans: Sequence[ReadableSpan]) -> SpanExportResult:
        for readable_span in spans:
            self.print_to_console(readable_span)
        return SpanExportResult.SUCCESS



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Any

import mcp.types as mcptypes
import uvicorn
from mcp.server import Server as MCPServer
from mcp.server.sse import SseServerTransport
from pydantic import BaseModel
from starlette.applications import Starlette
from starlette.responses import Response
from starlette.routing import Mount, Route

from any_agent.serving.server_handle import ServerHandle
from any_agent.utils import run_async_in_sync

if TYPE_CHECKING:
    from starlette.requests import Request

    from any_agent import AnyAgent


def _create_mcp_server_instance(agent: AnyAgent) -> MCPServer[Any]:
    server = MCPServer[Any]("any-agent-mcp-server")

    @server.list_tools()  # type: ignore[no-untyped-call,misc]
    async def handle_list_tools() -> list[mcptypes.Tool]:
        return [
            mcptypes.Tool(
                name=f"as-tool-{agent.config.name}",
                description=agent.config.description,
                inputSchema={
                    "type": "object",
                    "required": ["query"],
                    "properties": {
                        "query": {
                            "type": "string",
                            "description": "The prompt for the agent",
                        },
                    },
                },
            )
        ]

    @server.call_tool()  # type: ignore[misc]
    async def handle_call_tool(
        name: str, arguments: dict[str, Any]
    ) -> list[mcptypes.TextContent | mcptypes.ImageContent | mcptypes.EmbeddedResource]:
        result = await agent.run_async(arguments["query"])
        output = result.final_output
        if isinstance(output, BaseModel):
            serialized_output = output.model_dump_json()
        else:
            serialized_output = str(output)
        return [mcptypes.TextContent(type="text", text=serialized_output)]

    return server


def _get_mcp_app(
    agent: AnyAgent,
    endpoint: str,
) -> Starlette:
    """Provide an MCP server to be used in an event loop."""
    root = endpoint.lstrip("/").rstrip("/")
    msg_endpoint = f"/{root}/messages/"
    sse = SseServerTransport(msg_endpoint)
    server = _create_mcp_server_instance(agent)
    init_options = server.create_initialization_options()

    async def _handle_sse(request: Request):  # type: ignore[no-untyped-def]
        async with sse.connect_sse(
            request.scope, request.receive, request._send
        ) as streams:
            await server.run(streams[0], streams[1], init_options)
        # Return empty response to avoid NoneType error
        # Please check https://github.com/modelcontextprotocol/python-sdk/blob/1eb1bba83c70c3121bce7fc0263e5fac2c3f0520/src/mcp/server/sse.py#L33
        return Response()

    routes = [
        Route(f"/{root}/sse", endpoint=_handle_sse, methods=["GET"]),
        Mount(msg_endpoint, app=sse.handle_post_message),
    ]
    return Starlette(routes=routes)


async def serve_mcp_async(
    agent: AnyAgent,
    host: str,
    port: int,
    endpoint: str,
    log_level: str = "warning",
) -> ServerHandle:
    """Provide an MCP server to be used in an event loop."""
    config = uvicorn.Config(
        _get_mcp_app(agent, endpoint), host=host, port=port, log_level=log_level
    )
    uv_server = uvicorn.Server(config)
    task = asyncio.create_task(uv_server.serve())
    while not uv_server.started:  # noqa: ASYNC110
        await asyncio.sleep(0.1)
    return ServerHandle(task=task, server=uv_server)


def serve_mcp(
    agent: AnyAgent,
    host: str,
    port: int,
    endpoint: str,
    log_level: str = "warning",
) -> None:
    """Serve the MCP server."""

    # Note that the task should be kept somewhere
    # because the loop only keeps weak refs to tasks
    # https://docs.python.org/3/library/asyncio-task.html#asyncio.create_task
    async def run() -> None:
        server_handle = await serve_mcp_async(agent, host, port, endpoint, log_level)
        await server_handle.task

    return run_async_in_sync(run())



from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any

from opentelemetry.trace import Status, StatusCode

if TYPE_CHECKING:
    from opentelemetry.trace import Span


def _set_tool_output(tool_output: Any, span: Span) -> None:
    if tool_output is None:
        tool_output = "{}"

    status: Status | StatusCode = StatusCode.OK

    if isinstance(tool_output, str):
        try:
            json.loads(tool_output)
            output_type = "json"
        except json.decoder.JSONDecodeError:
            output_type = "text"
            if "Error calling tool:" in tool_output:
                status = Status(StatusCode.ERROR, description=tool_output)
    else:
        try:
            tool_output = json.dumps(tool_output, default=str, ensure_ascii=False)
            output_type = "json"
        except TypeError:
            tool_output = str(tool_output)
            output_type = "text"

    span.set_attributes(
        {"gen_ai.output": tool_output, "gen_ai.output.type": output_type}
    )
    span.set_status(status)



from abc import ABC, abstractmethod
from collections.abc import Sequence
from contextlib import suppress
from datetime import timedelta
from typing import Literal

from pydantic import Field, PrivateAttr

from any_agent.config import (
    AgentFramework,
    MCPSse,
    MCPStdio,
)
from any_agent.tools.mcp.mcp_connection import _MCPConnection
from any_agent.tools.mcp.mcp_server import _MCPServerBase

mcp_available = False
with suppress(ImportError):
    from agno.tools.mcp import MCPTools as AgnoMCPTools
    from mcp import ClientSession
    from mcp.client.sse import sse_client

    mcp_available = True


class AgnoMCPConnection(_MCPConnection["AgnoMCPTools"], ABC):
    _server: "AgnoMCPTools | None" = PrivateAttr(default=None)

    @abstractmethod
    async def list_tools(self) -> list["AgnoMCPTools"]:
        """List tools from the MCP server."""
        if self._server is None:
            msg = "MCP server is not set up. Please call `list_tools` from a concrete class."
            raise ValueError(msg)

        tools = await self._exit_stack.enter_async_context(self._server)
        return [tools]


class AgnoMCPStdioConnection(AgnoMCPConnection):
    mcp_tool: MCPStdio

    async def list_tools(self) -> list["AgnoMCPTools"]:
        """List tools from the MCP server."""
        from mcp import StdioServerParameters

        server_params = StdioServerParameters(
            command=self.mcp_tool.command,
            args=list(self.mcp_tool.args),
            env=self.mcp_tool.env,
        )
        kwargs = {}
        if self.mcp_tool.client_session_timeout_seconds:
            kwargs["timeout_seconds"] = int(
                self.mcp_tool.client_session_timeout_seconds
            )
        self._server = AgnoMCPTools(
            server_params=server_params,
            env=self.mcp_tool.env,
            include_tools=list(self.mcp_tool.tools) if self.mcp_tool.tools else None,
            **kwargs,  # type: ignore[arg-type]
        )
        return await super().list_tools()


class AgnoMCPSseConnection(AgnoMCPConnection):
    mcp_tool: MCPSse

    async def list_tools(self) -> list["AgnoMCPTools"]:
        """List tools from the MCP server."""
        client = sse_client(
            url=self.mcp_tool.url,
            headers=dict(self.mcp_tool.headers or {}),
        )
        sse_transport = await self._exit_stack.enter_async_context(client)
        stdio, write = sse_transport
        kwargs = {}
        if self.mcp_tool.client_session_timeout_seconds:
            kwargs["read_timeout_seconds"] = timedelta(
                seconds=self.mcp_tool.client_session_timeout_seconds
            )
        client_session = ClientSession(stdio, write, **kwargs)  # type: ignore[arg-type]
        session = await self._exit_stack.enter_async_context(client_session)
        await session.initialize()
        self._server = AgnoMCPTools(
            session=session,
            include_tools=list(self.mcp_tool.tools) if self.mcp_tool.tools else None,
        )
        return await super().list_tools()


class AgnoMCPServerBase(_MCPServerBase["AgnoMCPTools"], ABC):
    framework: Literal[AgentFramework.AGNO] = AgentFramework.AGNO
    tools: Sequence["AgnoMCPTools"] = Field(default_factory=list)

    def _check_dependencies(self) -> None:
        """Check if the required dependencies for the MCP server are available."""
        self.libraries = "any-agent[mcp,agno]"
        self.mcp_available = mcp_available
        super()._check_dependencies()


class AgnoMCPServerStdio(AgnoMCPServerBase):
    mcp_tool: MCPStdio

    async def _setup_tools(
        self, mcp_connection: _MCPConnection["AgnoMCPTools"] | None = None
    ) -> None:
        mcp_connection = mcp_connection or AgnoMCPStdioConnection(
            mcp_tool=self.mcp_tool
        )
        await super()._setup_tools(mcp_connection)


class AgnoMCPServerSse(AgnoMCPServerBase):
    mcp_tool: MCPSse

    async def _setup_tools(
        self, mcp_connection: _MCPConnection["AgnoMCPTools"] | None = None
    ) -> None:
        mcp_connection = mcp_connection or AgnoMCPSseConnection(mcp_tool=self.mcp_tool)
        await super()._setup_tools(mcp_connection)


AgnoMCPServer = AgnoMCPServerStdio | AgnoMCPServerSse



# mypy: disable-error-code="no-untyped-def,union-attr"
from __future__ import annotations

import json
from copy import deepcopy
from typing import TYPE_CHECKING, Any

from opentelemetry.trace import StatusCode

if TYPE_CHECKING:
    from collections.abc import Callable

    from opentelemetry.trace import Span
    from smolagents.models import ChatMessage

    from any_agent.frameworks.smolagents import SmolagentsAgent


from .common import _set_tool_output


def _set_llm_input(messages: list[ChatMessage], span: Span) -> None:
    if not messages:
        return
    span.set_attribute(
        "gen_ai.input.messages",
        json.dumps(
            [
                {
                    "role": message.role.value,  # type: ignore[attr-defined]
                    "content": message.content[0]["text"],  # type: ignore[index]
                }
                for message in messages
                if message.content
            ],
            default=str,
            ensure_ascii=False,
        ),
    )


def _set_llm_output(response: ChatMessage, span: Span) -> None:
    if content := response.content:
        span.set_attributes(
            {
                "gen_ai.output": str(content),
                "gen_ai.output.type": "text",
            }
        )
    if tool_calls := response.tool_calls:
        span.set_attributes(
            {
                "gen_ai.output": json.dumps(
                    [
                        {
                            "tool.name": tool_call.function.name,
                            "tool.args": tool_call.function.arguments,
                        }
                        for tool_call in tool_calls
                    ],
                    default=str,
                    ensure_ascii=False,
                ),
                "gen_ai.output.type": "json",
            }
        )

    if raw := response.raw:
        if token_usage := raw.get("usage", None):
            span.set_attributes(
                {
                    "gen_ai.usage.input_tokens": token_usage.prompt_tokens,
                    "gen_ai.usage.output_tokens": token_usage.completion_tokens,
                }
            )

        if response_model := raw.get("model", None):
            span.set_attribute("gen_ai.response.model", response_model)


class _SmolagentsInstrumentor:
    def __init__(self) -> None:
        self.first_llm_calls: set[int] = set()
        self._original_generate: Callable[..., Any] | None = None
        self._original_tools: Any | None = None

    def instrument(self, agent: SmolagentsAgent) -> None:
        if len(agent._running_traces) > 1:
            return

        tracer = agent._tracer

        self._original_generate = agent._agent.model.generate

        def wrap_generate(
            messages,
            stop_sequences=None,
            response_format=None,
            tools_to_call_from=None,
            **kwargs,
        ):
            model_id = str(agent._agent.model.model_id)
            with tracer.start_as_current_span(f"call_llm {model_id}") as span:
                span.set_attributes(
                    {
                        "gen_ai.operation.name": "call_llm",
                        "gen_ai.request.model": model_id,
                    }
                )
                trace_id = span.get_span_context().trace_id
                if trace_id not in self.first_llm_calls:
                    self.first_llm_calls.add(trace_id)
                    _set_llm_input(messages, span)

                response: ChatMessage = self._original_generate(  # type: ignore[misc]
                    messages,
                    stop_sequences,
                    response_format,
                    tools_to_call_from,
                    **kwargs,
                )

                _set_llm_output(response, span)

                span.set_status(StatusCode.OK)
                agent._running_traces[trace_id].add_span(span)

                return response

        agent._agent.model.generate = wrap_generate  # type: ignore[method-assign]

        class WrappedToolCall:
            def __init__(self, name, description, original_forward):
                self.name = name
                self.description = description
                self.original_forward = original_forward

            def forward(self, *args, **kwargs):
                with tracer.start_as_current_span(f"execute_tool {self.name}") as span:
                    span.set_attributes(
                        {
                            "gen_ai.operation.name": "execute_tool",
                            "gen_ai.tool.name": self.name,
                            "gen_ai.tool.description": self.description,
                            "gen_ai.tool.args": json.dumps(
                                kwargs, default=str, ensure_ascii=False
                            ),
                        }
                    )

                    output = self.original_forward(*args, **kwargs)
                    _set_tool_output(output, span)

                    trace_id = span.get_span_context().trace_id
                    agent._running_traces[trace_id].add_span(span)
                    return output

        self._original_tools = deepcopy(agent._agent.tools)
        wrapped_tools = {}
        for key, tool in agent._agent.tools.items():
            wrapped = WrappedToolCall(tool.name, tool.description, tool.forward)  # type: ignore[no-untyped-call]
            tool.forward = wrapped.forward
            wrapped_tools[key] = tool
        agent._agent.tools = wrapped_tools

    def uninstrument(self, agent: SmolagentsAgent) -> None:
        if len(agent._running_traces) > 1:
            return
        if self._original_generate is not None:
            agent._agent.model.generate = self._original_generate  # type: ignore[method-assign]
        if self._original_tools is not None:
            agent._agent.tools = self._original_tools



from datetime import datetime, timedelta
from typing import TYPE_CHECKING

from any_agent.logging import logger
from any_agent.tracing.agent_trace import AgentTrace

if TYPE_CHECKING:
    from any_agent.serving import A2AServingConfig


class TaskData:
    """Data stored for each task."""

    def __init__(self, task_id: str):
        """Initialize task data.

        Args:
            task_id: Unique identifier for the task

        """
        self.task_id = task_id
        self.agent_trace = AgentTrace()
        self.last_activity = datetime.now()
        self.created_at = datetime.now()

    def update_activity(self) -> None:
        """Update the last activity timestamp."""
        self.last_activity = datetime.now()

    def is_expired(self, timeout_minutes: int) -> bool:
        """Check if the task has expired.

        Args:
            timeout_minutes: Timeout in minutes

        Returns:
            True if task is expired, False otherwise

        """
        expiration_time = self.last_activity + timedelta(minutes=timeout_minutes)
        return datetime.now() > expiration_time


class TaskManager:
    """Manages agent conversation tasks for multi-turn interactions."""

    def __init__(self, config: "A2AServingConfig"):
        """Initialize the task manager.

        Args:
            config: Serving configuration containing task settings

        """
        self.config = config
        self._tasks: dict[str, TaskData] = {}
        self._last_cleanup: datetime | None = None

    def add_task(self, task_id: str) -> None:
        """Store a new task.

        This method will also trigger cleanup of expired tasks if it hasn't
        run recently (based on task_cleanup_interval_minutes).

        """
        self._cleanup_expired_tasks()

        self._tasks[task_id] = TaskData(task_id)
        logger.debug("Created new task: %s", task_id)

    def _get_task(self, task_id: str) -> TaskData | None:
        """Get task data by ID.

        Args:
            task_id: Task ID to retrieve

        Returns:
            TaskData if found and not expired, None otherwise

        """
        task = self._tasks.get(task_id)
        if not task:
            return None

        if task.is_expired(self.config.task_timeout_minutes):
            logger.debug("Task %s expired, removing", task_id)
            self._tasks.pop(task_id, None)
            return None

        task.update_activity()
        return task

    def update_task_trace(self, task_id: str, agent_trace: AgentTrace) -> None:
        """Update the agent trace for a task.

        Args:
            task_id: Task ID to update
            agent_trace: New agent trace to merge/store

        """
        task = self._get_task(task_id)
        if not task:
            logger.warning("Attempted to update non-existent task: %s", task_id)
            return

        task.agent_trace = agent_trace
        task.update_activity()

    def format_query_with_history(self, task_id: str, current_query: str) -> str:
        """Format a query with conversation history.

        Args:
            task_id: Task ID to get history for
            current_query: Current user query

        Returns:
            Formatted query string with history context

        """
        task = self._get_task(task_id)
        if not task:
            return current_query

        history = task.agent_trace.spans_to_messages()
        return self.config.history_formatter(history, current_query)

    def remove_task(self, task_id: str) -> None:
        """Remove a task.

        Args:
            task_id: Task ID to remove

        """
        if task_id in self._tasks:
            del self._tasks[task_id]
            logger.info("Removed task: %s", task_id)

    def _cleanup_expired_tasks(self) -> None:
        """Clean up expired tasks."""
        expired_tasks = []

        for task_id, task in self._tasks.items():
            if task.is_expired(self.config.task_timeout_minutes):
                expired_tasks.append(task_id)

        for task_id in expired_tasks:
            self.remove_task(task_id)

        # Update last cleanup time
        self._last_cleanup = datetime.now()

        if expired_tasks:
            logger.info("Cleaned up %d expired tasks", len(expired_tasks))
        else:
            logger.debug("No expired tasks to clean up")



from unittest.mock import MagicMock

from langchain_core.messages import BaseMessage
from langchain_core.outputs import LLMResult
from langchain_core.outputs.generation import Generation

from any_agent.tracing.instrumentation.langchain import (
    _LangChainInstrumentor,
    _set_llm_input,
    _set_llm_output,
)


def test_set_llm_input_missing_fields() -> None:
    """It should not fail when missing fields."""
    span = MagicMock()
    _set_llm_input([[BaseMessage(content="foo", type="human")]], span)

    span.set_attribute.assert_called_with(
        "gen_ai.input.messages", '[{"role": "user", "content": "foo"}]'
    )


def test_set_llm_output_missing_fields() -> None:
    """It should not fail when missing fields."""
    span = MagicMock()
    _set_llm_output(LLMResult(generations=[[Generation(text="")]]), span)

    span.set_attributes.assert_not_called()


def test_uninstrument_before_instrument() -> None:
    _LangChainInstrumentor().uninstrument(MagicMock())



# adapted from https://github.com/google/a2a-python/blob/main/examples/helloworld/test_client.py

import re
from collections.abc import Callable, Coroutine
from contextlib import suppress
from typing import TYPE_CHECKING, Any
from uuid import uuid4

from any_agent.utils.asyncio_sync import run_async_in_sync

if TYPE_CHECKING:
    from a2a.types import AgentCard

a2a_tool_available = False
with suppress(ImportError):
    import httpx
    from a2a.client import A2ACardResolver, A2AClient
    from a2a.types import (
        Message,
        MessageSendParams,
        Part,
        Role,
        SendMessageRequest,
        TextPart,
    )

    a2a_tool_available = True


async def a2a_tool_async(
    url: str, toolname: str | None = None, http_kwargs: dict[str, Any] | None = None
) -> Callable[[str], Coroutine[Any, Any, str]]:
    """Perform a query using A2A to another agent.

    Args:
        url (str): The url in which the A2A agent is located.
        toolname (str): The name for the created tool. Defaults to `call_{agent name in card}`.
            Leading and trailing whitespace are removed. Whitespace in the middle is replaced by `_`.
        http_kwargs (dict): Additional kwargs to pass to the httpx client.

    Returns:
        An async `Callable` that takes a query and returns the agent response.

    """
    if not a2a_tool_available:
        msg = "You need to `pip install 'any-agent[a2a]'` to use this tool"
        raise ImportError(msg)

    async with httpx.AsyncClient(follow_redirects=True) as resolver_client:
        a2a_agent_card: AgentCard = await (
            A2ACardResolver(httpx_client=resolver_client, base_url=url)
        ).get_agent_card()

    async def _send_query(query: str, task_id: str = str(uuid4())) -> str:
        async with httpx.AsyncClient(follow_redirects=True) as query_client:
            client = A2AClient(httpx_client=query_client, agent_card=a2a_agent_card)
            send_message_payload = SendMessageRequest(
                id=str(uuid4()),
                params=MessageSendParams(
                    message=Message(
                        role=Role.user,
                        parts=[Part(root=TextPart(text=query))],
                        # the id is not currently tracked
                        messageId=str(uuid4().hex),
                        taskId=task_id,
                    )
                ),
            )
            # TODO check how to capture exceptions and pass them on to the enclosing framework
            response = await client.send_message(
                send_message_payload, http_kwargs=http_kwargs
            )

            if not response.root:
                msg = (
                    "The A2A agent did not return a root. Are you using an A2A agent not managed by any-agent? "
                    "Please file an issue at https://github.com/mozilla-ai/any-agent/issues so we can help."
                )
                raise ValueError(msg)

            if hasattr(response.root, "error"):
                response_str = ""
                response_str += f"Error: {response.root.error.message}\n\n"
                response_str += f"Code: {response.root.error.code}\n\n"
                response_str += f"Data: {response.root.error.data}\n\n"
            elif hasattr(response.root, "result"):
                response_str = ""
                response_str += f"Status: {response.root.result.status.state}\n\n"
                response_str += f"""Message: {" ".join([part.root.text for part in response.root.result.status.message.parts if part.root.kind == "text"])}\n\n"""
                response_str += (
                    f"TaskId: {response.root.result.status.message.taskId}\n\n"
                )
                response_str += (
                    f"Timestamp: {response.root.result.status.timestamp}\n\n"
                )
            else:
                msg = (
                    "The A2A agent did not return a error or a result. Are you using an A2A agent not managed by any-agent? "
                    "Please file an issue at https://github.com/mozilla-ai/any-agent/issues so we can help."
                )
                raise ValueError(msg)

            return response_str

    new_name = toolname or a2a_agent_card.name
    new_name = re.sub(r"\s+", "_", new_name.strip())
    _send_query.__name__ = f"call_{new_name}"
    _send_query.__doc__ = f"""{a2a_agent_card.description}
        Send a query to the A2A hosted agent named {a2a_agent_card.name}.

        Agent description: {a2a_agent_card.description}

        Args:
            query (str): The query to perform.
            task_id (str): The task id to use for the conversation. Defaults to a new uuid. If you want to continue the conversation, you should provide the same task id that you received in a previous response.

        Returns:
            The result from the A2A agent, encoded as a json string.
    """
    return _send_query


def a2a_tool(
    url: str, toolname: str | None = None, http_kwargs: dict[str, Any] | None = None
) -> Callable[[str], str]:
    """Perform a query using A2A to another agent (synchronous version).

    Args:
        url (str): The url in which the A2A agent is located.
        toolname (str): The name for the created tool. Defaults to `call_{agent name in card}`.
            Leading and trailing whitespace are removed. Whitespace in the middle is replaced by `_`.
        http_kwargs (dict): Additional kwargs to pass to the httpx client.

    Returns:
        A sync `Callable` that takes a query and returns the agent response.

    """
    if not a2a_tool_available:
        msg = "You need to `pip install 'any-agent[a2a]'` to use this tool"
        raise ImportError(msg)

    # Fetch the async tool upfront to get proper name and documentation (otherwise the tool doesn't have the right name and documentation)
    async_tool = run_async_in_sync(a2a_tool_async(url, toolname, http_kwargs))

    def sync_wrapper(query: str) -> Any:
        """Execute the A2A tool query synchronously."""
        return run_async_in_sync(async_tool(query))

    # Copy essential metadata from the async tool
    sync_wrapper.__name__ = async_tool.__name__
    sync_wrapper.__doc__ = async_tool.__doc__

    return sync_wrapper



from typing import TYPE_CHECKING, override

from a2a.server.agent_execution import AgentExecutor, RequestContext
from a2a.server.events import EventQueue
from a2a.server.tasks import TaskUpdater
from a2a.types import Part, TextPart
from a2a.utils import (
    new_agent_parts_message,
    new_task,
)
from pydantic import BaseModel

from any_agent.logging import logger
from any_agent.serving.a2a.envelope import A2AEnvelope
from any_agent.serving.a2a.task_manager import TaskManager

if TYPE_CHECKING:
    from any_agent import AnyAgent


class AnyAgentExecutor(AgentExecutor):  # type: ignore[misc]
    """AnyAgentExecutor Implementation with task management for multi-turn conversations."""

    def __init__(self, agent: "AnyAgent", task_manager: TaskManager):
        """Initialize the AnyAgentExecutor.

        Args:
            agent: The agent to execute
            task_manager: Task manager to use for task management

        """
        self.agent = agent
        self.task_manager = task_manager

    @override
    async def execute(  # type: ignore[misc]
        self,
        context: RequestContext,
        event_queue: EventQueue,
    ) -> None:
        query = context.get_user_input()
        task = context.current_task

        # Extract or create task ID
        if not task:
            task = new_task(context.message)
            self.task_manager.add_task(task.id)
            await event_queue.enqueue_event(task)
        else:
            logger.debug("Task already exists: %s", task.model_dump_json(indent=2))

        formatted_query = self.task_manager.format_query_with_history(task.id, query)

        # This agent always produces Task objects.
        agent_trace = await self.agent.run_async(formatted_query)

        # Update task with new trace
        self.task_manager.update_task_trace(task.id, agent_trace)

        updater = TaskUpdater(event_queue, task.id, task.contextId)

        # Validate & interpret the envelope produced by the agent
        final_output = agent_trace.final_output

        if not isinstance(final_output, BaseModel):
            msg = f"Expected BaseModel, got {type(final_output)}, {final_output}"
            raise TypeError(msg)

        # Runtime attributes guaranteed by the dynamically created model.
        if not isinstance(final_output, A2AEnvelope):
            msg = "Final output must be an A2AEnvelope"
            raise TypeError(msg)

        task_status = final_output.task_status
        data_field = final_output.data

        # Convert payload to text we can stream to user
        if isinstance(data_field, BaseModel):
            result_text = data_field.model_dump_json()
        else:
            result_text = str(data_field)

        # Right now all task states will mark the state as final, because the server does not support streaming.
        # As we expand logic for streaming we may not need to always mark the state as final.
        await updater.update_status(
            task_status,
            message=new_agent_parts_message(
                [Part(root=TextPart(text=result_text))],
                task.contextId,
                task.id,
            ),
            final=True,
        )

    @override
    async def cancel(self, context: RequestContext, event_queue: EventQueue) -> None:  # type: ignore[misc]
        msg = "cancel not supported"
        raise ValueError(msg)



from abc import ABC, abstractmethod
from collections.abc import Sequence
from contextlib import AsyncExitStack
from textwrap import dedent
from typing import TYPE_CHECKING, Generic, Protocol, TypeVar, runtime_checkable

from pydantic import BaseModel, ConfigDict, PrivateAttr

from any_agent.config import MCPParams

if TYPE_CHECKING:
    from agents.mcp.server import MCPServer


@runtime_checkable
class HasName(Protocol):
    """Protocol for objects that have a name."""

    name: str


T = TypeVar("T")


class _MCPConnection(BaseModel, ABC, Generic[T]):
    mcp_tool: MCPParams
    _exit_stack: AsyncExitStack = PrivateAttr(default_factory=AsyncExitStack)

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @abstractmethod
    async def list_tools(self) -> list[T]:
        """List tools from the MCP server."""

    @property
    def server(self) -> "MCPServer | None":
        """Return the MCP server instance."""
        return None

    def _filter_tools(self, tools: Sequence[T]) -> Sequence[T]:
        """Filter the tools to only include the ones listed in mcp_tool['tools']."""
        requested_tools = list(self.mcp_tool.tools or [])

        if not requested_tools:
            return tools

        name_to_tool = {
            tool.name if isinstance(tool, HasName) else tool: tool for tool in tools
        }

        missing_tools = [name for name in requested_tools if name not in name_to_tool]
        if missing_tools:
            error_message = dedent(
                f"""Could not find all requested tools in the MCP server:
                Requested ({len(requested_tools)}): {requested_tools}
                Set ({len(name_to_tool)}):   {list(name_to_tool.keys())}
                Missing: {missing_tools}
            """
            )
            raise ValueError(error_message)

        return [name_to_tool[name] for name in requested_tools]



from typing import TYPE_CHECKING
from uuid import uuid4

from pydantic import BaseModel

from any_agent.config import AgentConfig, AgentFramework
from any_agent.tools.final_output import FinalOutputTool

from .any_agent import AnyAgent

try:
    from google.adk.agents.llm_agent import LlmAgent
    from google.adk.models.lite_llm import LiteLlm
    from google.adk.runners import InMemoryRunner
    from google.genai import types

    DEFAULT_MODEL_TYPE = LiteLlm
    adk_available = True
except ImportError:
    adk_available = False

if TYPE_CHECKING:
    from google.adk.models.base_llm import BaseLlm


class GoogleAgent(AnyAgent):
    """Google ADK agent implementation that handles both loading and running."""

    def __init__(self, config: AgentConfig):
        super().__init__(config)
        self._agent: LlmAgent | None = None

    @property
    def framework(self) -> AgentFramework:
        return AgentFramework.GOOGLE

    def _get_model(self, agent_config: AgentConfig) -> "BaseLlm":
        """Get the model configuration for a Google agent."""
        model_type = agent_config.model_type or DEFAULT_MODEL_TYPE
        model_args = agent_config.model_args or {}
        if self.config.output_type:
            model_args["tool_choice"] = "required"
        return model_type(
            model=agent_config.model_id,
            api_key=agent_config.api_key,
            api_base=agent_config.api_base,
            **model_args,
        )

    async def _load_agent(self) -> None:
        """Load the Google agent with the given configuration."""
        if not adk_available:
            msg = "You need to `pip install 'any-agent[google]'` to use this agent"
            raise ImportError(msg)

        tools, _ = await self._load_tools(self.config.tools)

        agent_type = self.config.agent_type or LlmAgent

        self._tools = tools

        instructions = self.config.instructions or ""
        if self.config.output_type:
            instructions += (
                "You must call the final_output tool when finished."
                "The 'answer' argument passed to the final_output tool must be a JSON string that matches the following schema:\n"
                f"{self.config.output_type.model_json_schema()}"
            )
            output_fn = FinalOutputTool(self.config.output_type)
            tools.append(output_fn)

        self._agent = agent_type(
            name=self.config.name,
            instruction=instructions,
            model=self._get_model(self.config),
            tools=tools,
            **self.config.agent_args or {},
            output_key="response",
        )

    async def _run_async(  # type: ignore[no-untyped-def]
        self,
        prompt: str,
        user_id: str | None = None,
        session_id: str | None = None,
        **kwargs,
    ) -> str | BaseModel:
        if not self._agent:
            error_message = "Agent not loaded. Call load_agent() first."
            raise ValueError(error_message)
        runner = InMemoryRunner(self._agent)
        user_id = user_id or str(uuid4())
        session_id = session_id or str(uuid4())
        await runner.session_service.create_session(
            app_name=runner.app_name,
            user_id=user_id,
            session_id=session_id,
        )

        if self.config.output_type:
            final_output = None
            final_output_attempts = 0
            # We allow for two retries: one to make it a proper json string, and one to make it a valid pydantic model
            max_output_attepts = 3

            async for event in runner.run_async(
                user_id=user_id,
                session_id=session_id,
                new_message=types.Content(role="user", parts=[types.Part(text=prompt)]),
                **kwargs,
            ):
                if not event.content or not event.content.parts:
                    continue

                # Check for final_output function responses
                for part in event.content.parts:
                    if (
                        part.function_response
                        and part.function_response.name == "final_output"
                        and part.function_response.response
                    ):
                        final_output_attempts += 1
                        if part.function_response.response.get("success"):
                            final_output = part.function_response.response.get("result")
                            break
                        if final_output_attempts >= max_output_attepts:
                            msg = f"Final output failed after {final_output_attempts} attempts"
                            raise ValueError(msg)

                if final_output or final_output_attempts >= max_output_attepts:
                    break

            if not final_output:
                msg = "No final response found"
                raise ValueError(msg)
            return self.config.output_type.model_validate_json(final_output)

        async for _ in runner.run_async(
            user_id=user_id,
            session_id=session_id,
            new_message=types.Content(role="user", parts=[types.Part(text=prompt)]),
            **kwargs,
        ):
            pass
        session = await runner.session_service.get_session(
            app_name=runner.app_name,
            user_id=user_id,
            session_id=session_id,
        )
        assert session, "Session should not be None"
        return str(session.state.get("response"))



import datetime

import pytest
from litellm.utils import validate_environment
from sse_starlette.sse import AppStatus

from any_agent import AgentConfig, AgentFramework, AnyAgent
from any_agent.config import MCPSse
from any_agent.serving import MCPServingConfig
from any_agent.tracing.agent_trace import AgentTrace
from tests.integration.helpers import wait_for_server_async


def _assert_valid_agent_trace(agent_trace: AgentTrace) -> None:
    """Assert that agent_trace is valid and has final output."""
    assert isinstance(agent_trace, AgentTrace)
    assert agent_trace.final_output


def _assert_contains_current_date_info(final_output: str) -> None:
    """Assert that the final output contains current date and time information."""
    now = datetime.datetime.now()
    assert all(
        [
            str(now.year) in final_output,
            str(now.day) in final_output,
            now.strftime("%B") in final_output,
        ]
    )


def _assert_has_tool_date_agent_call(agent_trace: AgentTrace) -> None:
    """Assert that the agent trace contains a tool execution span for the date agent."""
    assert any(
        span.is_tool_execution()
        and span.attributes.get("gen_ai.tool.name", None) == "as-tool-date_agent"
        for span in agent_trace.spans
    )


DATE_PROMPT = (
    "What date and time is it right now? "
    "In your answer please include the year, month, day, and time. "
    "Example answer could be something like 'Today is December 15, 2024'"
)


@pytest.mark.asyncio
async def test_mcp_serve(agent_framework: AgentFramework, test_port: int) -> None:
    """Tests that an agent contacts another as a tool using MCP.

    Note that there is an issue when using Google ADK: https://github.com/google/adk-python/pull/566
    """
    skip_reason = {
        AgentFramework.SMOLAGENTS: "async a2a is not supported",
    }
    if agent_framework in skip_reason:
        pytest.skip(
            f"Framework {agent_framework}, reason: {skip_reason[agent_framework]}"
        )
    kwargs = {}

    kwargs["model_id"] = "gpt-4.1-nano"
    agent_model = kwargs["model_id"]
    env_check = validate_environment(kwargs["model_id"])
    if not env_check["keys_in_environment"]:
        pytest.skip(f"{env_check['missing_keys']} needed for {agent_framework}")

    model_args = None

    main_agent = None
    server_handle = None

    try:
        tool_agent_endpoint = "tool_agent"

        # DATE AGENT

        import datetime

        def get_datetime() -> str:
            """Return the current date and time"""
            return str(datetime.datetime.now())

        date_agent_description = "Agent that can return the current date."
        date_agent_cfg = AgentConfig(
            instructions="Use the available tools to obtain additional information to answer the query.",
            name="date_agent",
            model_id=agent_model,
            description=date_agent_description,
            tools=[get_datetime],
            model_args=model_args,
        )
        date_agent = await AnyAgent.create_async(
            agent_framework=agent_framework,
            agent_config=date_agent_cfg,
        )

        # SERVING PROPER
        server_url = f"http://localhost:{test_port}/{tool_agent_endpoint}/sse"
        server_handle = await date_agent.serve_async(
            serving_config=MCPServingConfig(
                port=test_port,
                endpoint=f"/{tool_agent_endpoint}",
                log_level="info",
            )
        )
        ping_url = f"http://localhost:{test_port}"
        # We cannot use the SSE stream, as it will not be closed in the request
        await wait_for_server_async(ping_url)

        # Search agent is ready for card resolution
        main_agent_cfg = AgentConfig(
            instructions="Use the available tools to obtain additional information to answer the query.",
            description="The orchestrator that can use other agents via tools using the A2A protocol.",
            tools=[
                MCPSse(url=server_url, client_session_timeout_seconds=300),
            ],
            model_args=model_args,
            **kwargs,  # type: ignore[arg-type]
        )

        main_agent = await AnyAgent.create_async(
            agent_framework=agent_framework,
            agent_config=main_agent_cfg,
        )

        agent_trace = await main_agent.run_async(DATE_PROMPT)

        _assert_valid_agent_trace(agent_trace)
        _assert_contains_current_date_info(str(agent_trace.final_output))
        _assert_has_tool_date_agent_call(agent_trace)

    finally:
        if main_agent:
            if mcp_conn := main_agent._mcp_servers[0].mcp_connection:
                await mcp_conn._exit_stack.aclose()
        if AppStatus.should_exit_event is not None:
            AppStatus.should_exit_event.set()
            AppStatus.should_exit_event = None
        if server_handle:
            await server_handle.shutdown()



from __future__ import annotations

from typing import assert_never

from any_agent import AgentFramework

from .agno import _AgnoInstrumentor
from .google import _GoogleADKInstrumentor
from .langchain import _LangChainInstrumentor
from .llama_index import _LlamaIndexInstrumentor
from .openai import _OpenAIAgentsInstrumentor
from .smolagents import _SmolagentsInstrumentor
from .tinyagent import _TinyAgentInstrumentor


def _get_instrumentor_by_framework(
    framework: AgentFramework,
) -> (
    _AgnoInstrumentor
    | _GoogleADKInstrumentor
    | _LangChainInstrumentor
    | _LlamaIndexInstrumentor
    | _OpenAIAgentsInstrumentor
    | _SmolagentsInstrumentor
    | _TinyAgentInstrumentor
):
    if framework is AgentFramework.AGNO:
        return _AgnoInstrumentor()

    if framework is AgentFramework.GOOGLE:
        return _GoogleADKInstrumentor()

    if framework is AgentFramework.LANGCHAIN:
        return _LangChainInstrumentor()

    if framework is AgentFramework.LLAMA_INDEX:
        return _LlamaIndexInstrumentor()

    if framework is AgentFramework.OPENAI:
        return _OpenAIAgentsInstrumentor()

    if framework is AgentFramework.SMOLAGENTS:
        return _SmolagentsInstrumentor()

    if framework is AgentFramework.TINYAGENT:
        return _TinyAgentInstrumentor()

    assert_never(framework)



from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from any_agent import AgentConfig, AgentFramework, AnyAgent
from any_agent.frameworks.tinyagent import TinyAgent, ToolExecutor


async def sample_tool_function(arg1: int, arg2: str) -> str:
    assert isinstance(arg1, int), "arg1 should be an int"
    assert isinstance(arg2, str), "arg2 should be a str"
    return f"Received int: {arg1}, str: {arg2}"


@pytest.mark.asyncio
async def test_tool_argument_casting() -> None:
    agent: TinyAgent = await AnyAgent.create_async(
        AgentFramework.TINYAGENT, AgentConfig(model_id="gpt-4o")
    )  # type: ignore[assignment]

    # Register the sample tool function
    agent.clients["sample_tool"] = ToolExecutor(sample_tool_function)

    request = {
        "name": "sample_tool",
        "arguments": {
            "arg1": "42",  # This should be cast to int
            "arg2": 100,  # This should be cast to str
        },
    }

    # Call the tool and get the result
    result = await agent.clients["sample_tool"].call_tool(request)
    # Check the result
    assert result == "Received int: 42, str: 100"


def test_run_tinyagent_agent_custom_args() -> None:
    create_mock = MagicMock()
    agent_mock = AsyncMock()
    agent_mock.ainvoke.return_value = MagicMock()
    create_mock.return_value = agent_mock
    output = "The state capital of Pennsylvania is Harrisburg."

    agent = AnyAgent.create(AgentFramework.TINYAGENT, AgentConfig(model_id="gpt-4o"))
    with patch(
        "any_agent.frameworks.tinyagent.litellm.acompletion"
    ) as mock_acompletion:
        # Create a mock response object that properly mocks the LiteLLM response structure
        mock_response = MagicMock()
        mock_message = MagicMock()
        mock_message.content = output
        mock_message.tool_calls = []  # No tool calls in this response
        mock_message.model_dump.return_value = {
            "content": output,
            "role": "assistant",
            "tool_calls": None,
            "function_call": None,
            "annotations": [],
        }

        mock_response.choices = [MagicMock(message=mock_message)]

        # Make the acompletion function return this response
        mock_acompletion.return_value = mock_response

        # Call run which will eventually call _process_single_turn_with_tools
        result = agent.run("what's the state capital of Pennsylvania", debug=True)

        # Assert that the result contains the expected content
        assert output == result.final_output



from unittest.mock import MagicMock

from agno.models.message import Message

from any_agent.tracing.instrumentation.agno import (
    _AgnoInstrumentor,
    _set_llm_input,
    _set_llm_output,
)


def test_set_llm_input_missing_fields() -> None:
    """It should not fail when missing fields."""
    span = MagicMock()
    _set_llm_input([Message(role="user")], span)

    span.set_attribute.assert_called_with(
        "gen_ai.input.messages", '[{"role": "user", "content": null}]'
    )


def test_set_llm_output_missing_fields() -> None:
    """It should not fail when missing fields."""
    span = MagicMock()
    _set_llm_output(Message(role="assistant"), span)

    span.set_attributes.assert_called_once_with(
        {"gen_ai.usage.input_tokens": 0, "gen_ai.usage.output_tokens": 0}
    )


def test_uninstrument_before_instrument() -> None:
    _AgnoInstrumentor().uninstrument(MagicMock())



from .mcp.config_mcp import MCPServingConfig
from .mcp.server_mcp import (
    serve_mcp,
    serve_mcp_async,
)
from .server_handle import ServerHandle

__all__ = [
    "MCPServingConfig",
    "ServerHandle",
    "serve_mcp",
    "serve_mcp_async",
]

try:
    from .a2a.config_a2a import A2AServingConfig
    from .a2a.server_a2a import (
        _get_a2a_app,
        _get_a2a_app_async,
        serve_a2a,
        serve_a2a_async,
    )

    __all__ += [
        "A2AServingConfig",
        "_get_a2a_app",
        "_get_a2a_app_async",
        "serve_a2a",
        "serve_a2a_async",
    ]
except ImportError:
    msg = "You need to `pip install 'any-agent[a2a]'` to use this method."
    raise ImportError(msg) from None
